from __future__ import annotations

from collections.abc import Mapping
from typing import TYPE_CHECKING
from typing import Any

from hypothesis.extra.numpy import array_shapes
from hypothesis.strategies import DrawFn
from hypothesis.strategies import SearchStrategy
from hypothesis.strategies import composite
from xarray import DataArray

from utilities.xarray import DataArrayB
from utilities.xarray import DataArrayF
from utilities.xarray import DataArrayI
from utilities.xarray import DataArrayO

from .hypothesis import MaybeSearchStrategy  # noqa: TID252
from .hypothesis import lift_draw  # noqa: TID252
from .hypothesis import lists_fixed_length  # noqa: TID252
from .hypothesis import text_ascii  # noqa: TID252
from .numpy import bool_arrays  # noqa: TID252
from .numpy import float_arrays  # noqa: TID252
from .numpy import int_arrays  # noqa: TID252
from .numpy import str_arrays  # noqa: TID252
from .pandas import int_indexes  # noqa: TID252

if TYPE_CHECKING:  # pragma: no cover
    from utilities.pandas import IndexA
    from utilities.pandas import IndexI


@composite
def bool_data_arrays(
    _draw: DrawFn,
    indexes: MaybeSearchStrategy[Mapping[str, IndexA]] | None = None,
    /,
    *,
    fill: SearchStrategy[Any] | None = None,
    unique: MaybeSearchStrategy[bool] = False,
    name: MaybeSearchStrategy[str | None] = None,
    **indexes_kwargs: MaybeSearchStrategy[IndexA],
) -> DataArrayB:
    """Strategy for generating data arrays of booleans."""
    draw = lift_draw(_draw)
    indexes_ = draw(_merge_into_dict_of_indexes(indexes, **indexes_kwargs))
    shape = tuple(map(len, indexes_.values()))
    values = draw(bool_arrays(shape=shape, fill=fill, unique=unique))
    return DataArray(
        data=values, coords=indexes_, dims=list(indexes_), name=draw(name)
    )


@composite
def dicts_of_indexes(
    _draw: DrawFn,
    /,
    *,
    min_dims: int = 1,
    max_dims: int | None = None,
    min_side: int = 1,
    max_side: int | None = None,
) -> dict[str, IndexI]:
    """Strategy for generating dictionaries of indexes."""
    draw = lift_draw(_draw)
    shape = draw(
        array_shapes(
            min_dims=min_dims,
            max_dims=max_dims,
            min_side=min_side,
            max_side=max_side,
        )
    )
    ndims = len(shape)
    dims = draw(lists_fixed_length(text_ascii(), ndims, unique=True))
    indexes = (draw(int_indexes(n=length)) for length in shape)
    return dict(zip(dims, indexes, strict=True))


@composite
def float_data_arrays(
    _draw: DrawFn,
    indexes: MaybeSearchStrategy[Mapping[str, IndexA]] | None = None,
    /,
    *,
    min_value: MaybeSearchStrategy[float | None] = None,
    max_value: MaybeSearchStrategy[float | None] = None,
    allow_nan: MaybeSearchStrategy[bool] = False,
    allow_inf: MaybeSearchStrategy[bool] = False,
    allow_pos_inf: MaybeSearchStrategy[bool] = False,
    allow_neg_inf: MaybeSearchStrategy[bool] = False,
    integral: MaybeSearchStrategy[bool] = False,
    fill: SearchStrategy[Any] | None = None,
    unique: MaybeSearchStrategy[bool] = False,
    name: MaybeSearchStrategy[str | None] = None,
    **indexes_kwargs: MaybeSearchStrategy[IndexA],
) -> DataArrayF:
    """Strategy for generating data arrays of floats."""
    draw = lift_draw(_draw)
    indexes_ = draw(_merge_into_dict_of_indexes(indexes, **indexes_kwargs))
    shape = tuple(map(len, indexes_.values()))
    values = draw(
        float_arrays(
            shape=shape,
            min_value=min_value,
            max_value=max_value,
            allow_nan=allow_nan,
            allow_inf=allow_inf,
            allow_pos_inf=allow_pos_inf,
            allow_neg_inf=allow_neg_inf,
            integral=integral,
            fill=fill,
            unique=unique,
        )
    )
    return DataArray(
        data=values, coords=indexes_, dims=list(indexes_), name=draw(name)
    )


@composite
def int_data_arrays(
    _draw: DrawFn,
    indexes: MaybeSearchStrategy[Mapping[str, IndexA]] | None = None,
    /,
    *,
    min_value: MaybeSearchStrategy[int | None] = None,
    max_value: MaybeSearchStrategy[int | None] = None,
    fill: SearchStrategy[Any] | None = None,
    unique: MaybeSearchStrategy[bool] = False,
    name: MaybeSearchStrategy[str | None] = None,
    **indexes_kwargs: MaybeSearchStrategy[IndexA],
) -> DataArrayI:
    """Strategy for generating data arrays of ints."""
    draw = lift_draw(_draw)
    indexes_ = draw(_merge_into_dict_of_indexes(indexes, **indexes_kwargs))
    shape = tuple(map(len, indexes_.values()))
    values = draw(
        int_arrays(
            shape=shape,
            min_value=min_value,
            max_value=max_value,
            fill=fill,
            unique=unique,
        )
    )
    return DataArray(
        data=values, coords=indexes_, dims=list(indexes_), name=draw(name)
    )


@composite
def str_data_arrays(
    _draw: DrawFn,
    indexes: MaybeSearchStrategy[Mapping[str, IndexA]] | None = None,
    /,
    *,
    min_size: MaybeSearchStrategy[int] = 0,
    max_size: MaybeSearchStrategy[int | None] = None,
    allow_none: MaybeSearchStrategy[bool] = False,
    fill: SearchStrategy[Any] | None = None,
    unique: MaybeSearchStrategy[bool] = False,
    name: MaybeSearchStrategy[str | None] = None,
    **indexes_kwargs: MaybeSearchStrategy[IndexA],
) -> DataArrayO:
    """Strategy for generating data arrays of strings."""
    draw = lift_draw(_draw)
    indexes_ = draw(_merge_into_dict_of_indexes(indexes, **indexes_kwargs))
    shape = tuple(map(len, indexes_.values()))
    values = draw(
        str_arrays(
            shape=shape,
            min_size=min_size,
            max_size=max_size,
            allow_none=allow_none,
            fill=fill,
            unique=unique,
        )
    )
    return DataArray(
        data=values, coords=indexes_, dims=list(indexes_), name=draw(name)
    )


@composite
def _merge_into_dict_of_indexes(
    _draw: DrawFn,
    indexes: MaybeSearchStrategy[Mapping[str, IndexA]] | None = None,
    /,
    **indexes_kwargs: MaybeSearchStrategy[IndexA],
) -> dict[str, IndexA]:
    """Merge positional & kwargs of indexes into a dictionary."""
    draw = lift_draw(_draw)
    if (indexes is None) and (len(indexes_kwargs) == 0):
        return draw(dicts_of_indexes())
    indexes_out: dict[str, IndexA] = {}
    if indexes is not None:
        indexes_out |= dict(draw(indexes))
    indexes_out |= {k: draw(v) for k, v in indexes_kwargs.items()}
    return indexes_out



from __future__ import annotations

from collections.abc import Iterable
from collections.abc import Mapping
from collections.abc import Sequence
from typing import Any
from typing import Literal
from typing import cast
from typing import overload

from fastparquet import ParquetFile
from fastparquet import write
from pandas import DataFrame
from pandas import Series

from utilities.atomicwrites import writer
from utilities.errors import redirect_error
from utilities.iterables import is_iterable_not_str
from utilities.math import IntNonNeg
from utilities.numpy.numpy import datetime64ns
from utilities.numpy.numpy import has_dtype
from utilities.pandas import Int64
from utilities.pandas import check_range_index
from utilities.pandas import string
from utilities.pathlib import PathLike
from utilities.text import ensure_str

_Compression = Literal["gzip", ",snappy", "brotli", "lz4", "zstandard"]
Compression = _Compression | Mapping[str, _Compression | None]
_Op = Literal["==", "=", ">", ">=", "<", "<=", "!=", "in", "not in"]
_Filter = tuple[str, _Op, Any]
Filters = Sequence[_Filter] | Sequence[Sequence[_Filter]]
_PARQUET_DTYPES = {bool, cast(str, datetime64ns), float, Int64, string}


def count_rows(path: PathLike, /, *, filters: Filters | None = None) -> int:
    """Count the number of rows in a Parquet file."""
    return _get_parquet_file(path).count(
        filters=filters, **_maybe_row_filter(filters)
    )


def get_columns(path: PathLike, /) -> list[str]:
    """Get the columns in a Parquet file."""
    return _get_parquet_file(path).columns


def get_dtypes(path: PathLike, /) -> dict[str, Any]:
    """Get the dtypes in a Parquet file.

    Note that we store strings as categoricals, so we will report `string`
    instead of category here.
    """
    dtypes = _get_parquet_file(path).dtypes
    return {k: string if v == object else v for k, v in dtypes.items()}


def get_num_row_groups(path: PathLike, /) -> IntNonNeg:
    """Get the number of row groups in a Parquet file."""
    return len(_get_parquet_file(path).row_groups)


@overload
def read_parquet(
    path: PathLike,
    /,
    *,
    head: IntNonNeg | None = None,
    row_group: IntNonNeg | None = None,
    columns: str,
    filters: Filters | None = None,
) -> Series[Any]:
    ...


@overload
def read_parquet(
    path: PathLike,
    /,
    *,
    head: IntNonNeg | None = None,
    row_group: IntNonNeg | None = None,
    columns: list[str] | None = None,  # list, not Sequence
    filters: Filters | None = None,
) -> DataFrame:
    ...


def read_parquet(
    path: PathLike,
    /,
    *,
    head: IntNonNeg | None = None,
    row_group: IntNonNeg | None = None,
    columns: str | list[str] | None = None,  # list, not Sequence
    filters: Filters | None = None,
) -> Series[Any] | DataFrame:
    """Read a Parquet file into a Series/DataFrame."""
    file = _get_parquet_file(path, row_group=row_group)
    as_df = (columns is None) or is_iterable_not_str(columns)
    columns_use = columns if as_df else [columns]
    kwargs = _maybe_row_filter(filters)
    if head is None:
        df = file.to_pandas(columns=columns_use, filters=filters, **kwargs)
    else:
        df = file.head(head, columns=columns_use, filters=filters, **kwargs)
    dtypes = {k: string for k, v in df.items() if has_dtype(v, object)}
    df = df.astype(dtypes).reset_index(drop=True)
    return df if as_df else df[columns]


def _get_parquet_file(
    path: PathLike, /, *, row_group: IntNonNeg | None = None
) -> ParquetFile:
    """Read a Parquet file."""
    try:
        file = ParquetFile(path, verify=True)
    except TypeError as error:
        msg = f"{path=}"
        new = FileNotFoundError(msg)
        redirect_error(
            error, "argument of type 'PosixPath' is not iterable", new
        )
    if row_group is None:
        return file
    try:
        return file[row_group]
    except IndexError as error:
        msg = f"{path=}, {row_group=}"
        redirect_error(
            error, "list index out of range", InvalidRowGroupIndexError(msg)
        )


class InvalidRowGroupIndexError(IndexError):
    """Raised when a row group index is invalid."""


def _maybe_row_filter(filters: Filters | None, /) -> dict[str, bool]:
    """Add the `row_filter` argument if necessary."""
    return {} if filters is None else {"row_filter": True}


def write_parquet(
    df: DataFrame | Iterable[DataFrame],
    path: PathLike,
    /,
    *,
    extra_dtypes: Mapping[str, Any] | None = None,
    overwrite: bool = False,
    row_group_offsets: int | Sequence[int] | None = None,
    compression: Compression | None = "gzip",
) -> None:
    """Atomically write a DataFrame to a Parquet file."""
    with writer(path, overwrite=overwrite) as temp:
        if isinstance(df, DataFrame):
            _write_parquet_core(
                df,
                temp,
                extra_dtypes=extra_dtypes,
                row_group_offsets=row_group_offsets,
                compression=compression,
            )
        else:
            for i, df_i in enumerate(df):
                _write_parquet_core(
                    df_i,
                    temp,
                    extra_dtypes=extra_dtypes,
                    row_group_offsets=row_group_offsets,
                    compression=compression,
                    append=i >= 1,
                )


def _write_parquet_core(
    df: DataFrame,
    path: PathLike,
    /,
    *,
    extra_dtypes: Mapping[str, Any] | None = None,
    row_group_offsets: int | Sequence[int] | None = None,
    compression: Compression | None = "gzip",
    append: bool = False,
) -> None:
    """Atomically write a DataFrame to a Parquet file."""
    if len(df) == 0:
        msg = f"{df=}"
        raise EmptyDataFrameError(msg)
    check_range_index(df)
    for name, column in df.items():
        allowed_dtypes = _PARQUET_DTYPES
        if extra_dtypes is not None:
            try:
                extra_dtype = extra_dtypes[ensure_str(name)]
            except KeyError:
                pass
            else:
                allowed_dtypes = allowed_dtypes | {extra_dtype}
        if not has_dtype(column, allowed_dtypes):
            msg = f"{column=}"
            raise InvalidDTypeError(msg)
    write(
        path,
        df,
        row_group_offsets=row_group_offsets,
        compression=compression,
        append=append,
    )


class EmptyDataFrameError(ValueError):
    """Raised when a DataFrame is empty."""


class InvalidDTypeError(TypeError):
    """Raised when an invalid dtype is encountered."""



from __future__ import annotations

from collections.abc import Mapping
from collections.abc import Sequence
from collections.abc import Set as AbstractSet
from pathlib import Path
from typing import Any
from typing import Literal
from typing import cast

from fastparquet import write
from hypothesis import assume
from hypothesis import given
from hypothesis.extra.pandas import column
from hypothesis.extra.pandas import data_frames
from hypothesis.extra.pandas import range_indexes
from hypothesis.strategies import DataObject
from hypothesis.strategies import SearchStrategy
from hypothesis.strategies import booleans
from hypothesis.strategies import data
from hypothesis.strategies import dictionaries
from hypothesis.strategies import floats
from hypothesis.strategies import integers
from hypothesis.strategies import lists
from hypothesis.strategies import none
from hypothesis.strategies import sampled_from
from hypothesis.strategies import sets
from hypothesis.strategies import tuples
from numpy import float32
from numpy import nan
from pandas import DataFrame
from pandas import RangeIndex
from pandas import Series
from pandas import concat
from pandas.testing import assert_frame_equal
from pandas.testing import assert_series_equal
from pytest import mark
from pytest import param
from pytest import raises

from utilities.fastparquet import _PARQUET_DTYPES
from utilities.fastparquet import EmptyDataFrameError
from utilities.fastparquet import InvalidDTypeError
from utilities.fastparquet import InvalidRowGroupIndexError
from utilities.fastparquet import _get_parquet_file
from utilities.fastparquet import count_rows
from utilities.fastparquet import get_columns
from utilities.fastparquet import get_dtypes
from utilities.fastparquet import get_num_row_groups
from utilities.fastparquet import read_parquet
from utilities.fastparquet import write_parquet
from utilities.hypothesis import dates_pd
from utilities.hypothesis import lists_fixed_length
from utilities.hypothesis import temp_paths
from utilities.hypothesis import text_ascii
from utilities.numpy import datetime64ns
from utilities.pandas import DataFrameRangeIndexError
from utilities.pandas import Int64
from utilities.pandas import string


class TestCountRows:
    @given(rows=lists(booleans(), min_size=1), root=temp_paths())
    def test_main(self, rows: Sequence[bool], root: Path) -> None:
        n = len(rows)
        df = DataFrame(rows, index=RangeIndex(n), columns=["value"])
        write_parquet(df, path := root.joinpath("df.parq"))
        result = count_rows(path)
        assert result == n


class TestGetColumns:
    @given(columns=sets(text_ascii()), root=temp_paths())
    def test_main(self, columns: AbstractSet[str], root: Path) -> None:
        as_list = list(columns)
        df = DataFrame(nan, index=RangeIndex(1), columns=as_list, dtype=float)
        write_parquet(df, path := root.joinpath("df.parq"))
        result = get_columns(path)
        assert result == as_list


class TestGetDtypes:
    @given(
        dtypes=dictionaries(
            text_ascii(), sampled_from(sorted(_PARQUET_DTYPES, key=repr))
        ),
        root=temp_paths(),
    )
    def test_main(self, dtypes: Mapping[str, Any], root: Path) -> None:
        df = DataFrame(None, index=RangeIndex(1), columns=list(dtypes)).astype(
            dtypes
        )
        write_parquet(df, path := root.joinpath("df.parq"))
        result = get_dtypes(path)
        assert result == dtypes


class TestGetNumRowGroups:
    def test_main(self, tmp_path: Path) -> None:
        df = DataFrame(0.0, index=RangeIndex(2), columns=["value"])
        write_parquet(
            df, path := tmp_path.joinpath("df.parq"), row_group_offsets=1
        )
        result = get_num_row_groups(path)
        expected = 2
        assert result == expected


class TestGetParquetFile:
    def test_main(self, tmp_path: Path) -> None:
        path = tmp_path.joinpath("file")
        df = DataFrame(0.0, index=RangeIndex(1), columns=["value"])
        write(path, df)
        _ = _get_parquet_file(path)

    def test_get_row_group(self, tmp_path: Path) -> None:
        path = tmp_path.joinpath("file")
        df = DataFrame(0.0, index=RangeIndex(2), columns=["value"])
        write(path, df, row_group_offsets=1)
        _ = _get_parquet_file(path, row_group=0)
        _ = _get_parquet_file(path, row_group=1)
        with raises(InvalidRowGroupIndexError):
            _ = _get_parquet_file(path, row_group=2)

    @mark.parametrize("as_str", [param(True), param(False)])
    def test_error(self, *, tmp_path: Path, as_str: bool) -> None:
        path = tmp_path.joinpath("file")
        path_use = path.as_posix() if as_str else path
        with raises(FileNotFoundError):
            _ = _get_parquet_file(path_use)


class TestReadAndWriteParquet:
    @given(data=data(), root=temp_paths(), as_series=booleans())
    @mark.parametrize(
        ("elements", "dtype"),
        [
            param(booleans(), bool),
            param(dates_pd() | none(), datetime64ns),
            param(floats(-10.0, 10.0) | none(), float),
            param(integers(-10, 10) | none(), Int64),
            param(text_ascii() | none(), string),
        ],
    )
    def test_writing_df(
        self,
        *,
        data: DataObject,
        elements: SearchStrategy[Any],
        dtype: Any,
        root: Path,
        as_series: bool,
    ) -> None:
        rows = data.draw(lists(elements, min_size=1))
        n = len(rows)
        df = DataFrame(rows, index=RangeIndex(n), columns=["value"]).astype(
            dtype
        )
        write_parquet(df, path := root.joinpath("df.parq"))
        head = data.draw(sampled_from([n, None]))
        columns = "value" if as_series else None
        read = read_parquet(path, head=head, columns=columns)
        if as_series:
            assert isinstance(read, Series)
            assert_series_equal(read, df["value"])
        else:
            assert isinstance(read, DataFrame)
            assert_frame_equal(read, df)

    @given(value1=floats(), value2=floats(), root=temp_paths())
    def test_writing_iterable_of_dfs(
        self, value1: float, value2: float, root: Path
    ) -> None:
        df = DataFrame(
            [[value1], [value2]], index=RangeIndex(2), columns=["value"]
        )
        parts = [df.iloc[:1], df.iloc[1:].reset_index(drop=True)]
        write_parquet(parts, path := root.joinpath("df.parq"))
        result = read_parquet(path)
        assert_frame_equal(result, df)

    @given(
        data=data(),
        column1=text_ascii(),
        column2=text_ascii(),
        root=temp_paths(),
    )
    def test_series_from_dataframe_with_two_string_columns(
        self, data: DataObject, column1: str, column2: str, root: Path
    ) -> None:
        _ = assume(column1 != column2)
        elements = text_ascii() | none()
        rows = data.draw(lists(tuples(elements, elements), min_size=1))
        df = DataFrame(
            rows,
            index=RangeIndex(len(rows)),
            columns=[column1, column2],
            dtype=string,
        )
        write_parquet(df, path := root.joinpath("df.parq"))
        sr = read_parquet(path, columns=column1)
        assert_series_equal(sr, df[column1])

    @given(value1=floats(), value2=floats(), root=temp_paths())
    def test_read_row_groups(
        self, value1: float, value2: float, root: Path
    ) -> None:
        df = DataFrame(
            [[value1], [value2]], index=RangeIndex(2), columns=["value"]
        )
        write_parquet(df, path := root.joinpath("df.parq"), row_group_offsets=1)
        result1 = read_parquet(path, row_group=0)
        expected1 = df.iloc[:1]
        assert_frame_equal(result1, expected1)
        result2 = read_parquet(path, row_group=1)
        expected2 = df.iloc[1:].reset_index(drop=True)
        assert_frame_equal(result2, expected2)
        with raises(InvalidRowGroupIndexError):
            _ = read_parquet(path, row_group=2)

    @given(
        data=data(),
        root=temp_paths(),
        num_dfs=integers(1, 10),
        row_group_offsets=integers(1, 10),
    )
    def test_iterable_of_dfs_with_strings(
        self, data: DataObject, root: Path, num_dfs: int, row_group_offsets: int
    ) -> None:
        def as_str(df: DataFrame, /) -> DataFrame:
            return df.astype(cast(Literal["string"], string))

        elements = data_frames(
            [cast(Any, column)("value", elements=text_ascii())],
            index=range_indexes(min_size=1),
        ).map(as_str)
        dfs = data.draw(lists_fixed_length(elements, num_dfs))
        write_parquet(
            dfs,
            path := root.joinpath("df.parq"),
            row_group_offsets=row_group_offsets,
        )
        result = read_parquet(path)
        expected = concat(dfs).reset_index(drop=True)
        assert_frame_equal(result, expected)


class TestWriteParquet:
    @given(value=text_ascii(), root=temp_paths())
    def test_strings_are_stored_as_objects(
        self, value: str, root: Path
    ) -> None:
        df = DataFrame(
            value, index=RangeIndex(1), columns=["value"], dtype=string
        )
        write_parquet(df, path := root.joinpath("df.parq"))
        file = _get_parquet_file(path)
        dtypes = file.dtypes
        expected = {"value": object}
        assert dtypes == expected

    def test_empty_dataframe_error(self, tmp_path: Path) -> None:
        df = DataFrame()
        with raises(EmptyDataFrameError):
            write_parquet(df, tmp_path.joinpath("df.parq"))

    def test_check_range_index(self, tmp_path: Path) -> None:
        df = DataFrame(nan, index=[0], columns=["value"])
        with raises(DataFrameRangeIndexError):
            write_parquet(df, tmp_path.joinpath("df.parq"))

    def test_check_invalid_dtype(self, tmp_path: Path) -> None:
        df = DataFrame(
            nan, index=RangeIndex(1), columns=["value"], dtype=float32
        )
        with raises(InvalidDTypeError):
            write_parquet(df, tmp_path.joinpath("df.parq"))

    def test_extra_dtype(self, tmp_path: Path) -> None:
        df = DataFrame(
            nan, index=RangeIndex(1), columns=["value"], dtype=float32
        )
        write_parquet(
            df, tmp_path.joinpath("df.parq"), extra_dtypes={"value": float32}
        )

    def test_extra_dtype_ignored(self, tmp_path: Path) -> None:
        df = DataFrame(
            nan, index=RangeIndex(1), columns=["value"], dtype=float32
        )
        with raises(InvalidDTypeError):
            write_parquet(
                df,
                tmp_path.joinpath("df.parq"),
                extra_dtypes={"other": float32},
            )



import datetime as dt
from enum import Enum
from enum import auto
from pathlib import Path
from typing import Any
from typing import Literal

from hypothesis import given
from hypothesis.strategies import integers
from luigi import BoolParameter
from luigi import FloatParameter
from luigi import IntParameter
from luigi import ListParameter
from luigi import OptionalBoolParameter
from luigi import OptionalFloatParameter
from luigi import OptionalIntParameter
from luigi import OptionalListParameter
from luigi import OptionalPathParameter
from luigi import OptionalStrParameter
from luigi import Parameter
from luigi import PathParameter
from luigi import Task
from pytest import mark
from pytest import param
from pytest import raises
from sqlalchemy import Engine
from typed_settings import settings

from utilities.datetime import TODAY
from utilities.hypothesis import namespace_mixins
from utilities.luigi import AmbiguousDateError
from utilities.luigi import AmbiguousDatetimeError
from utilities.luigi import DateHourParameter
from utilities.luigi import DateMinuteParameter
from utilities.luigi import DateParameter
from utilities.luigi import DateSecondParameter
from utilities.luigi import EngineParameter
from utilities.luigi import EnumParameter
from utilities.luigi import InvalidAnnotationAndKeywordsError
from utilities.luigi import InvalidAnnotationError
from utilities.luigi import TimeParameter
from utilities.luigi import WeekdayParameter
from utilities.luigi import build_params_mixin
from utilities.luigi.attrs import _map_annotation
from utilities.luigi.attrs import _map_date_annotation
from utilities.luigi.attrs import _map_datetime_annotation
from utilities.luigi.attrs import _map_iterable_annotation
from utilities.luigi.attrs import _map_keywords
from utilities.luigi.attrs import _map_union_annotation
from utilities.sentinel import Sentinel


class TestBuildParamsMixin:
    @given(namespace_mixin=namespace_mixins())
    def test_no_field(self, namespace_mixin: Any) -> None:
        @settings
        class Config:
            value: int = 0

        config = Config()
        Params = build_params_mixin(config)  # noqa: N806

        class Example(namespace_mixin, Params, Task):
            pass

        task = Example()
        assert task.value == 0

    @given(namespace_mixin=namespace_mixins())
    def test_with_field(self, namespace_mixin: Any) -> None:
        @settings
        class Config:
            date: dt.date = TODAY

        config = Config()
        Params = build_params_mixin(config, date="date")  # noqa: N806

        class Example(namespace_mixin, Params, Task):
            pass

        task = Example()
        assert task.date == TODAY


class TestMapAnnotation:
    @mark.parametrize(
        ("ann", "expected"),
        [
            param(bool, BoolParameter),
            param(Engine, EngineParameter),
            param(dt.time, TimeParameter),
            param(float, FloatParameter),
            param(int, IntParameter),
            param(Path, PathParameter),
            param(str, Parameter),
            param(frozenset[bool], ListParameter),
            param(list[bool], ListParameter),
            param(set[bool], ListParameter),
            param(bool | None, OptionalBoolParameter),
            param(frozenset[bool] | None, OptionalListParameter),
            param(list[bool] | None, OptionalListParameter),
            param(set[bool] | None, OptionalListParameter),
        ],
    )
    def test_main(self, ann: Any, expected: type[Parameter]) -> None:
        result = _map_annotation(ann)
        param = result()
        assert isinstance(param, expected)

    @mark.parametrize("kind", [param("date"), param("weekday")])
    def test_date_success(self, kind: Literal["date", "weekday"]) -> None:
        _ = _map_annotation(dt.date, date=kind)

    def test_date_error(self) -> None:
        with raises(AmbiguousDateError):
            _ = _map_annotation(dt.date)

    @mark.parametrize("kind", [param("hour"), param("minute"), param("second")])
    def test_datetime_success(
        self, kind: Literal["hour", "minute", "second"]
    ) -> None:
        _ = _map_annotation(dt.datetime, datetime=kind)

    def test_datetime_error(self) -> None:
        with raises(AmbiguousDatetimeError):
            _ = _map_annotation(dt.datetime)

    def test_enum(self) -> None:
        class Example(Enum):
            member = auto()

        result = _map_annotation(Example)
        param = result()
        assert isinstance(param, EnumParameter)
        assert param._enum is Example  # noqa: SLF001

    @mark.parametrize("ann", [param(None), param(Sentinel)])
    def test_invalid(self, ann: Any) -> None:
        with raises(InvalidAnnotationError):
            _ = _map_annotation(ann)


class TestMapDateAnnotation:
    @mark.parametrize(
        ("kind", "expected"),
        [param("date", DateParameter), param("weekday", WeekdayParameter)],
    )
    def test_main(
        self, kind: Literal["date", "weekday"], expected: type[Parameter]
    ) -> None:
        result = _map_date_annotation(kind=kind)
        param = result()
        assert isinstance(param, expected)


class TestMapDatetimeAnnotation:
    @given(interval=integers(1, 10))
    @mark.parametrize(
        ("kind", "expected"),
        [
            param("hour", DateHourParameter),
            param("minute", DateMinuteParameter),
            param("second", DateSecondParameter),
        ],
    )
    def test_main(
        self,
        kind: Literal["hour", "minute", "second"],
        interval: int,
        expected: type[Parameter],
    ) -> None:
        result = _map_datetime_annotation(kind=kind, interval=interval)
        param = result()
        assert isinstance(param, expected)


class TestMapIterableAnnotation:
    @mark.parametrize(
        "ann", [param(frozenset[bool]), param(list[bool]), param(set[bool])]
    )
    def test_main(self, ann: Any) -> None:
        assert _map_iterable_annotation(ann) is ListParameter

    @mark.parametrize("ann", [param(None), param(bool), param(bool | None)])
    def test_invalid(self, ann: Any) -> None:
        with raises(InvalidAnnotationError):
            _ = _map_iterable_annotation(ann)


class TestMapKeywords:
    @mark.parametrize("kind", [param("date"), param("weekday")])
    def test_date(self, kind: str) -> None:
        result = _map_keywords(dt.date, kind)
        expected = {"date": kind}
        assert result == expected

    @mark.parametrize("kind", [param("hour"), param("minute"), param("second")])
    def test_datetime_kind_only(self, kind: str) -> None:
        result = _map_keywords(dt.datetime, kind)
        expected = {"datetime": kind}
        assert result == expected

    @given(interval=integers(1, 10))
    @mark.parametrize("kind", [param("hour"), param("minute"), param("second")])
    def test_datetime_kind_and_interval(self, interval: int, kind: str) -> None:
        result = _map_keywords(dt.datetime, (kind, interval))
        expected = {"datetime": kind, "interval": interval}
        assert result == expected

    @mark.parametrize(
        ("ann", "kwargs"),
        [
            param(None, None),
            param(bool, None),
            param(dt.date, "invalid"),
            param(dt.datetime, "invalid"),
            param(dt.datetime, (0,)),
            param(dt.datetime, (0, 1)),
            param(dt.datetime, (0, 1, 2)),
        ],
    )
    def test_invalid(self, ann: Any, kwargs: Any) -> None:
        with raises(InvalidAnnotationAndKeywordsError):
            _ = _map_keywords(ann, kwargs)


class TestMapUnionAnnotation:
    @mark.parametrize(
        ("ann", "expected"),
        [
            param(bool | None, OptionalBoolParameter),
            param(float | None, OptionalFloatParameter),
            param(Path | None, OptionalPathParameter),
            param(int | None, OptionalIntParameter),
            param(str | None, OptionalStrParameter),
            param(list[bool] | None, OptionalListParameter),
        ],
    )
    def test_main(self, ann: Any, expected: type[Parameter]) -> None:
        result = _map_union_annotation(ann)
        param = result()
        assert isinstance(param, expected)

    @mark.parametrize(
        "ann", [param(list[bool]), param(Sentinel | None), param(int | float)]
    )
    def test_invalid(self, ann: Any) -> None:
        with raises(InvalidAnnotationError):
            _ = _map_union_annotation(ann)

    def test_invalid_enum(self) -> None:
        class Example(Enum):
            member = auto()

        with raises(InvalidAnnotationError):
            _ = _map_union_annotation(Example | None)



from __future__ import annotations

from holoviews import Curve
from hypothesis import given
from pytest import mark

from utilities.fpdf2 import yield_pdf
from utilities.hypothesis import text_ascii
from utilities.platform import SYSTEM
from utilities.platform import System


class TestYieldPDF:
    @given(text=text_ascii(min_size=1))
    def test_add_fixed_width_text(self, text: str) -> None:
        with yield_pdf() as pdf:
            pdf.add_fixed_width_text(text)

    @mark.skipif(SYSTEM is not System.linux, reason="Linux only")
    def test_add_plot(self) -> None:
        curve = Curve([])
        with yield_pdf() as pdf:
            pdf.add_plot(curve)

    @given(header=text_ascii(min_size=1))
    def test_header(self, header: str) -> None:
        with yield_pdf(header=header) as pdf:
            pdf.header()

    def test_footer(self) -> None:
        with yield_pdf() as pdf:
            pdf.footer()



from __future__ import annotations

from .luigi import AwaitTask
from .luigi import AwaitTime
from .luigi import DateHourParameter
from .luigi import DateMinuteParameter
from .luigi import DateParameter
from .luigi import DateSecondParameter
from .luigi import EnumParameter
from .luigi import ExternalFile
from .luigi import ExternalTask
from .luigi import PathTarget
from .luigi import TimeParameter
from .luigi import WeekdayParameter
from .luigi import build
from .luigi import clone
from .luigi import get_dependencies_downstream
from .luigi import get_dependencies_upstream
from .luigi import get_task_classes

__all__ = [
    "AwaitTask",
    "AwaitTime",
    "build",
    "clone",
    "DateHourParameter",
    "DateMinuteParameter",
    "DateParameter",
    "DateSecondParameter",
    "EnumParameter",
    "ExternalFile",
    "ExternalTask",
    "get_dependencies_downstream",
    "get_dependencies_upstream",
    "get_task_classes",
    "PathTarget",
    "TimeParameter",
    "WeekdayParameter",
]


try:
    from .attrs import AmbiguousDateError
    from .attrs import AmbiguousDatetimeError
    from .attrs import InvalidAnnotationAndKeywordsError
    from .attrs import InvalidAnnotationError
    from .attrs import build_params_mixin
except ModuleNotFoundError:  # pragma: no cover
    pass
else:
    __all__ += [
        "AmbiguousDateError",
        "AmbiguousDatetimeError",
        "build_params_mixin",
        "InvalidAnnotationAndKeywordsError",
        "InvalidAnnotationError",
    ]


try:
    from .semver import VersionParameter
except ModuleNotFoundError:  # pragma: no cover
    pass
else:
    __all__ += ["VersionParameter"]


try:
    from .sqlalchemy import DatabaseTarget
    from .sqlalchemy import EngineParameter
    from .sqlalchemy import TableParameter
except ModuleNotFoundError:  # pragma: no cover
    pass
else:
    __all__ += ["DatabaseTarget", "EngineParameter", "TableParameter"]



from __future__ import annotations

from typing import Any
from typing import cast

from hypothesis.strategies import DrawFn
from hypothesis.strategies import composite
from sqlalchemy import Engine
from sqlalchemy import MetaData

from utilities.sqlalchemy import create_engine

from .hypothesis import temp_paths  # noqa: TID252


@composite
def sqlite_engines(
    _draw: DrawFn, /, *, metadata: MetaData | None = None, base: Any = None
) -> Engine:
    """Strategy for generating SQLite engines."""
    temp_path = _draw(temp_paths())
    path = temp_path.joinpath("db.sqlite")
    engine = create_engine("sqlite", database=path.as_posix())
    if metadata is not None:
        metadata.create_all(engine)
    if base is not None:
        base.metadata.create_all(engine)

    # attach temp_path to the engine, so as to keep it alive
    cast(Any, engine).temp_path = temp_path

    return engine



from __future__ import annotations

from re import compile


def extract_group(pattern: str, text: str, /) -> str:
    """Extract a group.

    The regex must have 1 capture group, and this must match exactly once.
    """
    if compile(pattern).groups <= 1:
        (result,) = extract_groups(pattern, text)
        return result
    raise MultipleCaptureGroupsError(pattern)


class MultipleCaptureGroupsError(ValueError):
    """Raised when multiple capture groups are found."""


def extract_groups(pattern: str, text: str, /) -> list[str]:
    """Extract multiple groups.

    The regex may have any number of capture groups, and they must collectively
    match exactly once.
    """
    compiled = compile(pattern)
    if (n_groups := compiled.groups) == 0:
        raise NoCaptureGroupsError(pattern)
    results = compiled.findall(text)
    msg = f"{pattern=}, {text=}"
    if (n_results := len(results)) == 0:
        raise NoMatchesError(msg)
    if n_results == 1:
        if n_groups == 1:
            return results
        (result,) = results
        return list(result)
    raise MultipleMatchesError(msg)


class NoCaptureGroupsError(ValueError):
    """Raised when no capture groups are found."""


class NoMatchesError(ValueError):
    """Raised when no matches are found."""


class MultipleMatchesError(ValueError):
    """Raised when multiple matches are found."""



from __future__ import annotations

from subprocess import run

from hypothesis.strategies import DrawFn
from hypothesis.strategies import composite

from utilities.pathlib import temp_cwd
from utilities.tempfile import TemporaryDirectory

from .hypothesis import MaybeSearchStrategy  # noqa: TID252
from .hypothesis import lift_draw  # noqa: TID252
from .hypothesis import temp_dirs  # noqa: TID252


@composite
def git_repos(
    _draw: DrawFn, /, *, branch: MaybeSearchStrategy[str | None] = None
) -> TemporaryDirectory:
    draw = lift_draw(_draw)
    temp_dir = draw(temp_dirs())
    path = temp_dir.path
    with temp_cwd(path):
        _ = run(["git", "init"], check=True)  # noqa: S603, S607
        _ = run(
            ["git", "config", "user.name", "User"],  # noqa: S603, S607
            check=True,
        )
        _ = run(
            ["git", "config", "user.email", "a@z.com"],  # noqa: S603, S607
            check=True,
        )
        file = path.joinpath("file")
        file.touch()
        file_str = file.as_posix()
        _ = run(["git", "add", file_str], check=True)  # noqa: S603, S607
        _ = run(["git", "commit", "-m", "add"], check=True)  # noqa: S603, S607
        _ = run(["git", "rm", file_str], check=True)  # noqa: S603, S607
        _ = run(["git", "commit", "-m", "rm"], check=True)  # noqa: S603, S607
        if (branch := draw(branch)) is not None:
            _ = run(
                ["git", "checkout", "-b", branch],  # noqa: S603, S607
                check=True,
            )
    return temp_dir



from __future__ import annotations

from beartype.door import die_if_unbearable
from beartype.roar import BeartypeAbbyHintViolation
from pytest import mark
from pytest import param
from pytest import raises

from utilities.types import Number
from utilities.typing import IterableStrs
from utilities.typing import NeverError
from utilities.typing import SequenceStrs
from utilities.typing import never


class TestIterableStrs:
    @mark.parametrize(
        "x",
        [
            param(["a", "b", "c"]),
            param(("a", "b", "c")),
            param({"a", "b", "c"}),
            param({"a": 1, "b": 2, "c": 3}),
        ],
    )
    def test_pass(self, x: IterableStrs) -> None:
        die_if_unbearable(x, IterableStrs)

    def test_fail(self) -> None:
        with raises(BeartypeAbbyHintViolation):
            die_if_unbearable("abc", SequenceStrs)


class TestSequenceStrs:
    @mark.parametrize("x", [param(["a", "b", "c"]), param(("a", "b", "c"))])
    def test_pass(self, x: SequenceStrs) -> None:
        die_if_unbearable(x, SequenceStrs)

    @mark.parametrize(
        "x",
        [param({"a", "b", "c"}), param({"a": 1, "b": 2, "c": 3}), param("abc")],
    )
    def test_fail(self, x: IterableStrs | str) -> None:
        with raises(BeartypeAbbyHintViolation):
            die_if_unbearable(x, SequenceStrs)


class TestNever:
    def test_main(self) -> None:
        with raises(NeverError):
            never(None)  # type: ignore


class TestNumber:
    @mark.parametrize("x", [param(0), param(0.0)])
    def test_main(self, x: Number) -> None:
        die_if_unbearable(x, Number)



from __future__ import annotations

from re import escape

from pytest import raises

from utilities.more_itertools import EmptyIterableError
from utilities.more_itertools import MultipleElementsError
from utilities.more_itertools import one


class TestOne:
    def test_empty(self) -> None:
        with raises(
            EmptyIterableError,
            match=escape("too few items in iterable (expected 1)"),
        ):
            _ = one([])

    def test_one(self) -> None:
        assert one([None]) is None

    def test_multiple(self) -> None:
        with raises(
            MultipleElementsError,
            match="Expected exactly one item in iterable, but got 1, 2, and "
            "perhaps more",
        ):
            _ = one([1, 2, 3])



from __future__ import annotations

from typing import Any

from hypothesis import given
from hypothesis.strategies import DataObject
from hypothesis.strategies import data
from hypothesis.strategies import integers
from hypothesis_sqlalchemy.sample import table_records_lists
from luigi import Task
from sqlalchemy import Column
from sqlalchemy import Engine
from sqlalchemy import Integer
from sqlalchemy import MetaData
from sqlalchemy import Table
from sqlalchemy import insert
from sqlalchemy import select
from sqlalchemy.orm import declarative_base

from utilities.hypothesis import namespace_mixins
from utilities.hypothesis import sqlite_engines
from utilities.luigi import DatabaseTarget
from utilities.luigi import EngineParameter
from utilities.luigi import TableParameter


class TestDatabaseTarget:
    @given(data=data(), engine=sqlite_engines())
    def test_main(self, data: DataObject, engine: Engine) -> None:
        table = Table(
            "example",
            MetaData(),
            Column("id1", Integer, primary_key=True),
            Column("id2", Integer, primary_key=True),
        )
        sel = select(table).where(table.c.id1 == 0)
        target = DatabaseTarget(sel, engine)
        assert not target.exists()
        rows = data.draw(
            table_records_lists(
                table, id1=integers(0, 10), min_size=1, max_size=10
            )
        )
        with engine.begin() as conn:
            table.create(conn)
            _ = conn.execute(insert(table).values(rows))
        expected = any(row[0] == 0 for row in rows)
        assert target.exists() is expected


class TestEngineParameter:
    @given(engine=sqlite_engines())
    def test_main(self, engine: Engine) -> None:
        param = EngineParameter()
        norm = param.normalize(engine)
        new_engine = param.parse(param.serialize(norm))
        assert new_engine.url == norm.url


class TestTableParameter:
    @given(namespace_mixin=namespace_mixins())
    def test_main(self, namespace_mixin: Any) -> None:
        class ExampleTask(namespace_mixin, Task):
            table = TableParameter()

        class ExampleTable(declarative_base()):
            __tablename__ = "example"

            id_ = Column(Integer, primary_key=True)

        _ = ExampleTask(ExampleTable)



from __future__ import annotations

from hypothesis import given
from hypothesis.strategies import floats
from hypothesis.strategies import integers
from pytest import raises

from utilities.holoviews import ArrayNameIsEmptyStringError
from utilities.holoviews import ArrayNameNotAStringError
from utilities.holoviews import plot_curve
from utilities.hypothesis import assume_does_not_raise
from utilities.hypothesis import bool_data_arrays
from utilities.hypothesis import float_data_arrays
from utilities.hypothesis import int_indexes
from utilities.hypothesis import text_ascii
from utilities.xarray import DataArrayB1
from utilities.xarray import DataArrayF1


class TestPlotCurve:
    @given(
        array=float_data_arrays(dim=int_indexes(), name=text_ascii(min_size=1))
    )
    def test_main(self, array: DataArrayF1) -> None:
        curve = plot_curve(array)
        assert curve.kdims == ["dim"]
        assert curve.vdims == [array.name]
        assert curve.label == array.name

    @given(
        array=float_data_arrays(dim=int_indexes(), name=text_ascii(min_size=1)),
        label=text_ascii(min_size=1),
    )
    def test_label(self, array: DataArrayF1, label: str) -> None:
        curve = plot_curve(array, label=label)
        assert curve.label == label

    @given(
        array=float_data_arrays(dim=int_indexes(), name=text_ascii(min_size=1)),
        smooth=integers(1, 10),
    )
    def test_smooth(self, array: DataArrayF1, smooth: int) -> None:
        with assume_does_not_raise(RuntimeWarning):
            curve = plot_curve(array, smooth=smooth)
        assert curve.label == f"{array.name} (MA{smooth})"

    @given(
        array=float_data_arrays(dim=int_indexes(), name=text_ascii(min_size=1)),
        aspect=floats(1.0, 10.0),
    )
    def test_aspect(self, array: DataArrayF1, aspect: float) -> None:
        _ = plot_curve(array, aspect=aspect)

    @given(array=float_data_arrays(dim=int_indexes()))
    def test_array_name_not_a_string(self, array: DataArrayF1) -> None:
        with raises(ArrayNameNotAStringError):
            _ = plot_curve(array)

    @given(
        array=float_data_arrays(dim=int_indexes(), name=text_ascii(max_size=0))
    )
    def test_array_name_is_empty_string(self, array: DataArrayF1) -> None:
        with raises(ArrayNameIsEmptyStringError):
            _ = plot_curve(array)

    @given(
        array=bool_data_arrays(dim=int_indexes(), name=text_ascii(min_size=1))
    )
    def test_boolean(self, array: DataArrayB1) -> None:
        _ = plot_curve(array)



from __future__ import annotations

from subprocess import PIPE
from subprocess import check_output

from hypothesis import given
from hypothesis.strategies import DataObject
from hypothesis.strategies import data
from hypothesis.strategies import none

from utilities.git import _GET_BRANCH_NAME
from utilities.hypothesis import git_repos
from utilities.hypothesis import text_ascii
from utilities.tempfile import TemporaryDirectory


class TestGitRepos:
    @given(data=data())
    def test_fixed(self, *, data: DataObject) -> None:
        branch = data.draw(text_ascii(min_size=1) | none())
        repo = data.draw(git_repos(branch=branch))
        assert isinstance(repo, TemporaryDirectory)
        path = repo.path
        assert set(path.iterdir()) == {path.joinpath(".git")}
        if branch is not None:
            output = check_output(
                _GET_BRANCH_NAME, stderr=PIPE, cwd=path, text=True  # noqa: S603
            )
            assert output.strip("\n") == branch



from __future__ import annotations

from collections.abc import Mapping
from collections.abc import Set as AbstractSet
from typing import Any
from typing import NoReturn

SequenceStrs = list[str] | tuple[str, ...]
IterableStrs = SequenceStrs | AbstractSet[str] | Mapping[str, Any]
Number = float | int


def never(x: NoReturn, /) -> NoReturn:
    """Never return. Used for exhaustive pattern matching."""
    msg = f'"never" was run with {x}'
    raise NeverError(msg)


class NeverError(Exception):
    """Raised when `never` is run."""



from __future__ import annotations

from typing import Any
from typing import TypeVar
from typing import cast

from holoviews import save

from utilities.atomicwrites import writer
from utilities.pathlib import PathLike

_T = TypeVar("_T")


def apply_opts(plot: _T, /, **opts: Any) -> _T:
    """Apply a set of options to a plot."""
    return cast(Any, plot).opts(**opts)


def relabel_plot(plot: _T, label: str, /) -> _T:
    """Re-label a plot."""
    return cast(Any, plot).relabel(label)


def save_plot(plot: Any, path: PathLike, /, *, overwrite: bool = False) -> None:
    """Atomically save a plot to disk."""
    with writer(path, overwrite=overwrite) as temp:  # pragma: os-ne-linux
        save(plot, temp, backend="bokeh")



from __future__ import annotations

from os import getenv
from typing import TypedDict
from typing import cast

from hypothesis import Verbosity
from hypothesis import settings


class Kwargs(TypedDict, total=False):
    deadline: None
    print_blob: bool
    report_multiple_bugs: bool


kwargs = cast(
    Kwargs,
    {"deadline": None, "print_blob": True, "report_multiple_bugs": False},
)
settings.register_profile("default", max_examples=100, **kwargs)
settings.register_profile("dev", max_examples=10, **kwargs)
settings.register_profile("ci", max_examples=1000, **kwargs)
settings.register_profile(
    "debug", max_examples=10, verbosity=Verbosity.verbose, **kwargs
)
settings.load_profile(getenv("HYPOTHESIS_PROFILE", "default"))



from __future__ import annotations

from inspect import signature
from pathlib import Path
from time import sleep
from typing import Any

from pytest import MonkeyPatch
from pytest import mark
from pytest import param

from utilities.pytest import is_pytest
from utilities.pytest import throttle
from utilities.text import strip_and_dedent
from utilities.typing import IterableStrs


class TestPytestOptions:
    def test_unknown_mark(self, testdir: Any) -> None:
        testdir.makepyfile(
            """
            from pytest import mark

            @mark.unknown
            def test_main():
                assert True
            """
        )
        result = testdir.runpytest()
        result.assert_outcomes(errors=1)
        result.stdout.re_match_lines([r".*Unknown pytest\.mark\.unknown"])

    def test_configured_mark_unknown_option(self, testdir: Any) -> None:
        testdir.makeconftest(
            """
            from utilities.pytest import add_pytest_configure

            def pytest_configure(config):
                add_pytest_configure(config, [("slow", "slow to run")])
            """
        )
        testdir.makepyfile(
            """
            from pytest import mark

            @mark.slow
            def test_main():
                assert True
            """
        )
        result = testdir.runpytest("--slow")
        result.stderr.re_match_lines(
            ["-c: error: unrecognized arguments: --slow"]
        )

    @mark.parametrize(
        ("case", "passed", "skipped", "matches"),
        [param([], 0, 1, [".*3: pass --slow"]), param(["--slow"], 1, 0, [])],
    )
    def test_configured_one_mark_and_option(
        self,
        testdir: Any,
        case: IterableStrs,
        passed: int,
        skipped: int,
        matches: IterableStrs,
    ) -> None:
        testdir.makeconftest(
            """
            from utilities.pytest import add_pytest_addoption
            from utilities.pytest import add_pytest_collection_modifyitems
            from utilities.pytest import add_pytest_configure

            def pytest_addoption(parser):
                add_pytest_addoption(parser, ["slow"])

            def pytest_collection_modifyitems(config, items):
                add_pytest_collection_modifyitems(config, items, ["slow"])

            def pytest_configure(config):
                add_pytest_configure(config, [("slow", "slow to run")])
            """
        )
        testdir.makepyfile(
            """
            from pytest import mark

            @mark.slow
            def test_main():
                assert True
            """
        )
        result = testdir.runpytest("-rs", *case)
        result.assert_outcomes(passed=passed, skipped=skipped)
        result.stdout.re_match_lines(matches)

    @mark.parametrize(
        ("case", "passed", "skipped", "matches"),
        [
            param(
                [],
                1,
                3,
                [
                    "SKIPPED.*: pass --slow",
                    "SKIPPED.*: pass --fast",
                    "SKIPPED.*: pass --slow --fast",
                ],
            ),
            param(
                ["--slow"],
                2,
                2,
                ["SKIPPED.*: pass --fast", "SKIPPED.*: pass --slow --fast"],
            ),
            param(
                ["--fast"],
                2,
                2,
                ["SKIPPED.*: pass --slow", "SKIPPED.*: pass --slow --fast"],
            ),
            param(["--slow", "--fast"], 4, 0, []),
        ],
    )
    def test_configured_two_marks_and_options(
        self,
        testdir: Any,
        case: IterableStrs,
        passed: int,
        skipped: int,
        matches: IterableStrs,
    ) -> None:
        testdir.makeconftest(
            """
            from utilities.pytest import add_pytest_addoption
            from utilities.pytest import add_pytest_collection_modifyitems
            from utilities.pytest import add_pytest_configure

            def pytest_addoption(parser):
                add_pytest_addoption(parser, ["slow", "fast"])

            def pytest_collection_modifyitems(config, items):
                add_pytest_collection_modifyitems(
                    config, items, ["slow", "fast"],
                )

            def pytest_configure(config):
                add_pytest_configure(
                    config, [("slow", "slow to run"), ("fast", "fast to run")],
                )
            """
        )
        testdir.makepyfile(
            """
            from pytest import mark

            def test_none():
                assert True

            @mark.slow
            def test_slow():
                assert True

            @mark.fast
            def test_fast():
                assert True

            @mark.slow
            @mark.fast
            def test_both():
                assert True
            """
        )
        result = testdir.runpytest("-rs", *case, "--randomly-dont-reorganize")
        result.assert_outcomes(passed=passed, skipped=skipped)
        result.stdout.re_match_lines(matches)


class TestIsPytest:
    def test_function(self) -> None:
        assert is_pytest()

    def test_disable(self, monkeypatch: MonkeyPatch) -> None:
        monkeypatch.delenv("PYTEST_CURRENT_TEST")
        assert not is_pytest()


class TestThrottle:
    def test_main(self, testdir: Any, tmp_path: Path) -> None:
        root_str = tmp_path.as_posix()
        contents = f"""
            from utilities.pytest import throttle

            @throttle(root={root_str!r}, duration=1.0)
            def test_main():
                assert True
            """
        testdir.makepyfile(strip_and_dedent(contents))

        result = testdir.runpytest()
        result.assert_outcomes(passed=1)

        result = testdir.runpytest()
        result.assert_outcomes(skipped=1)

        sleep(1.0)
        result = testdir.runpytest()
        result.assert_outcomes(passed=1)

    def test_signature(self) -> None:
        @throttle()
        def func(*, fix: bool) -> None:
            assert fix

        def other(*, fix: bool) -> None:
            assert fix

        assert signature(func) == signature(other)



#!/usr/bin/env python
from __future__ import annotations

from logging import info
from pathlib import Path
from subprocess import run
from sys import path

from tomli import loads

path.insert(0, Path("src").resolve().as_posix())

from utilities.logging import basic_config  # noqa: E402

basic_config()


with Path("pyproject.toml").open() as fh:
    contents = fh.read()
groups: set[str] = set(loads(contents)["project"]["optional-dependencies"]) - {
    "dev",
    "test",
}


for group in groups:
    info("Compiling group %r...", group)
    requirements = f"requirements/{group}.txt"
    _ = run(
        [  # noqa: S603, S607
            "pip-compile",
            "--allow-unsafe",
            f"--extra={group}",
            "--extra=test",
            f"--output-file={requirements}",
            "--quiet",
            "--upgrade",
        ],
        check=True,
    )
    _ = run(["pip-sync", requirements], check=True)  # noqa: S603, S607
    module = f"test_{group.replace('-', '_')}.py"
    path = Path("src", "tests", module).as_posix()
    _ = run(["pytest", "--no-cov", path], check=True)  # noqa: S603, S607



from __future__ import annotations

from typing import Any

from luigi import Parameter
from luigi import Target
from sqlalchemy import Engine
from sqlalchemy import Select
from sqlalchemy import create_engine
from sqlalchemy.exc import DatabaseError
from sqlalchemy.exc import NoSuchTableError
from typing_extensions import override

from utilities.sqlalchemy import get_table_name
from utilities.sqlalchemy import redirect_to_no_such_table_error


class DatabaseTarget(Target):
    """A target point to a set of rows in a database."""

    def __init__(self, sel: Select[Any], engine: Engine, /) -> None:
        super().__init__()
        self._sel = sel.limit(1)
        self._engine = engine

    def exists(self) -> bool:  # type: ignore
        try:
            with self._engine.begin() as conn:
                res = conn.execute(self._sel).one_or_none()
        except DatabaseError as error:
            try:
                redirect_to_no_such_table_error(self._engine, error)
            except NoSuchTableError:
                return False
        else:
            return res is not None


class EngineParameter(Parameter):
    """Parameter taking the value of a SQLAlchemy engine."""

    @override
    def normalize(self, x: Engine) -> Engine:
        """Normalize an `Engine` argument."""
        return x

    @override
    def parse(self, x: str) -> Engine:
        """Parse an `Engine` argument."""
        return create_engine(x)

    @override
    def serialize(self, x: Engine) -> str:
        """Serialize an `Engine` argument."""
        return x.url.render_as_string()


class TableParameter(Parameter):
    """Parameter taking the value of a SQLAlchemy table."""

    @override
    def normalize(self, x: Any) -> Any:
        """Normalize a `Table` or model argument."""
        return x

    @override
    def serialize(self, x: Any) -> str:
        """Serialize a `Table` or model argument."""
        return get_table_name(x)



from __future__ import annotations

import datetime as dt
from collections.abc import Callable
from contextlib import suppress
from enum import Enum
from functools import partial
from pathlib import Path
from types import UnionType
from types import new_class
from typing import Any
from typing import Literal
from typing import TypeVar
from typing import cast
from typing import get_args
from typing import get_origin

from attrs import asdict
from attrs import fields
from luigi import BoolParameter
from luigi import FloatParameter
from luigi import IntParameter
from luigi import ListParameter
from luigi import OptionalBoolParameter
from luigi import OptionalFloatParameter
from luigi import OptionalIntParameter
from luigi import OptionalListParameter
from luigi import OptionalPathParameter
from luigi import OptionalStrParameter
from luigi import Parameter
from luigi import PathParameter

from utilities.class_name import get_class_name
from utilities.types import NoneType
from utilities.typing import never

from .luigi import DateHourParameter  # noqa: TID252
from .luigi import DateMinuteParameter  # noqa: TID252
from .luigi import DateParameter  # noqa: TID252
from .luigi import DateSecondParameter  # noqa: TID252
from .luigi import EnumParameter  # noqa: TID252
from .luigi import TimeParameter  # noqa: TID252
from .luigi import WeekdayParameter  # noqa: TID252

_T = TypeVar("_T")


def build_params_mixin(obj: _T, /, **kwargs: Any) -> type[_T]:
    """Build a mixin of parameters for use in a `Task`."""
    mapping = asdict(obj)

    def exec_body(namespace: dict[str, Any], /) -> None:
        for field in fields(type(obj)):
            key = field.name
            ann = field.type
            try:
                value = kwargs[key]
            except KeyError:
                kwargs_ann = {}
            else:
                kwargs_ann = _map_keywords(ann, value)
            param_cls = _map_annotation(ann, **kwargs_ann)
            namespace[key] = param_cls(default=mapping[key], positional=False)

    name = get_class_name(obj)
    return cast(type[_T], new_class(f"{name}Params", exec_body=exec_body))


def _map_keywords(ann: Any, kwargs: Any, /) -> dict[str, Any]:
    """Map an annotation and a set of keywords to a dictionary."""
    msg = f"{ann=}, {kwargs=}"
    if not isinstance(ann, type):
        raise InvalidAnnotationAndKeywordsError(msg)
    if issubclass(ann, dt.datetime):
        allowed = {"hour", "minute", "second"}
        if isinstance(kwargs, str) and (kwargs in allowed):
            return {"datetime": kwargs}
        if isinstance(kwargs, tuple):
            try:
                datetime, interval = kwargs
            except ValueError:
                raise InvalidAnnotationAndKeywordsError(msg) from None
            if (
                isinstance(datetime, str)
                and (datetime in allowed)
                and isinstance(interval, int)
            ):
                return {"datetime": datetime, "interval": interval}
        raise InvalidAnnotationAndKeywordsError(msg)
    if (
        issubclass(ann, dt.date)
        and isinstance(kwargs, str)
        and (kwargs in {"date", "weekday"})
    ):
        return {"date": kwargs}
    raise InvalidAnnotationAndKeywordsError(msg)


class InvalidAnnotationAndKeywordsError(Exception):
    """Raised when an (annotation, keywords) pair is invalid."""


def _map_annotation(  # noqa: PLR0911, PLR0912
    ann: Any,
    /,
    *,
    date: Literal["date", "weekday"] | None = None,
    datetime: Literal["hour", "minute", "second"] | None = None,
    interval: int = 1,
) -> type[Parameter] | Callable[..., Parameter]:
    """Map an annotation to a parameter class."""
    with suppress(InvalidAnnotationError):
        return _map_iterable_annotation(ann)
    with suppress(InvalidAnnotationError):
        return _map_union_annotation(ann)
    msg = f"{ann=}"
    if not isinstance(ann, type):
        raise InvalidAnnotationError(msg) from None
    if issubclass(ann, bool):
        return BoolParameter
    if issubclass(ann, dt.datetime):
        if datetime is None:
            raise AmbiguousDatetimeError
        return _map_datetime_annotation(kind=datetime, interval=interval)
    if issubclass(ann, dt.date):
        if date is None:
            raise AmbiguousDateError
        return _map_date_annotation(kind=date)
    if issubclass(ann, dt.time):
        return TimeParameter
    if issubclass(ann, Enum):
        return partial(EnumParameter, ann)
    if issubclass(ann, float):
        return FloatParameter
    if issubclass(ann, int):
        return IntParameter
    if issubclass(ann, Path):
        return PathParameter
    if issubclass(ann, str):
        return Parameter
    try:
        from sqlalchemy import Engine

        from utilities.luigi.sqlalchemy import EngineParameter
    except ModuleNotFoundError:  # pragma: no cover
        pass
    else:
        if issubclass(ann, Engine):
            return EngineParameter
    raise InvalidAnnotationError(msg)


class InvalidAnnotationError(Exception):
    """Raised when an annotation is invalid."""


def _map_iterable_annotation(ann: Any, /) -> type[ListParameter]:
    """Map an iterable annotation to a parameter class."""
    if get_origin(ann) in {frozenset, list, set}:
        return ListParameter
    msg = f"{ann=}"
    raise InvalidAnnotationError(msg)


def _map_union_annotation(
    ann: Any, /
) -> type[Parameter] | Callable[..., Parameter]:
    """Map a union annotation to a parameter class."""
    msg = f"{ann=}"
    if get_origin(ann) is not UnionType:
        raise InvalidAnnotationError(msg)
    args = [arg for arg in get_args(ann) if arg is not NoneType]
    try:
        (arg,) = args
    except ValueError:
        raise InvalidAnnotationError(msg) from None
    if (inner := _map_annotation(arg)) is BoolParameter:
        return OptionalBoolParameter
    if inner is FloatParameter:
        return OptionalFloatParameter
    if inner is IntParameter:
        return OptionalIntParameter
    if inner is ListParameter:
        return OptionalListParameter
    if inner is PathParameter:
        return OptionalPathParameter
    if inner is Parameter:
        return OptionalStrParameter
    raise InvalidAnnotationError(msg)


def _map_date_annotation(
    *, kind: Literal["date", "weekday"]
) -> type[Parameter] | Callable[..., Parameter]:
    """Map a date annotation to a parameter class."""
    if kind == "date":
        return DateParameter
    if kind == "weekday":
        return WeekdayParameter
    return never(kind)  # pragma: no cover


class AmbiguousDateError(Exception):
    """Raised when a date is ambiguous."""


def _map_datetime_annotation(
    *, kind: Literal["hour", "minute", "second"], interval: int = 1
) -> type[Parameter] | Callable[..., Parameter]:
    """Map a datetime annotation to a parameter class."""
    if kind == "hour":
        cls = DateHourParameter
    elif kind == "minute":
        cls = DateMinuteParameter
    elif kind == "second":
        cls = DateSecondParameter
    else:
        return never(kind)  # pragma: no cover
    return partial(cls, interval=interval)


class AmbiguousDatetimeError(Exception):
    """Raised when a datetime is ambiguous."""



from __future__ import annotations

from typing import Any

from pytest import mark
from pytest import param

from utilities.hashlib import md5_hash


class TestMD5Hash:
    @mark.parametrize(
        ("obj", "expected"),
        [
            param(b"", "d41d8cd98f00b204e9800998ecf8427e"),
            param(b"bytes", "4b3a6218bb3e3a7303e8a171a60fcf92"),
            param("", "d41d8cd98f00b204e9800998ecf8427e"),
            param("text", "1cb251ec0d568de6a929b520c4aed8d1"),
            param(None, "7d1e55650014e21b4568ed3e3d1fc531"),
        ],
    )
    def test_main(self, obj: Any, expected: str) -> None:
        assert md5_hash(obj) == expected



from __future__ import annotations

from numpy import apply_along_axis
from numpy import clip
from numpy import full_like
from numpy import isnan
from numpy import nan
from numpy import zeros_like
from scipy.stats import norm

from utilities.math import FloatFinNonNeg
from utilities.numpy import NDArrayF
from utilities.numpy import NDArrayF1
from utilities.numpy import is_zero


def ppf(
    array: NDArrayF, cutoff: FloatFinNonNeg, /, *, axis: int = -1
) -> NDArrayF:
    """Apply the PPF transform to an array of data."""
    return apply_along_axis(_ppf_1d, axis, array, cutoff)


def _ppf_1d(array: NDArrayF1, cutoff: FloatFinNonNeg, /) -> NDArrayF1:
    if (i := isnan(array)).all():
        return array
    if i.any():
        j = ~i
        out = full_like(array, nan, dtype=float)
        out[j] = _ppf_1d(array[j], cutoff)
        return out
    low, high = min(array), max(array)
    if is_zero(span := high - low):
        return zeros_like(array, dtype=float)
    centred = (array - low) / span
    phi = norm.cdf(-cutoff)
    ppf = norm.ppf((1.0 - 2.0 * phi) * centred + phi)
    return clip(ppf, a_min=-cutoff, a_max=cutoff)



from __future__ import annotations

from typing import Annotated
from typing import TypeAlias

from xarray import DataArray

from utilities.numpy import DTypeB
from utilities.numpy import DTypeDns
from utilities.numpy import DTypeF
from utilities.numpy import DTypeI
from utilities.numpy import DTypeO
from utilities.numpy import NDim0
from utilities.numpy import NDim1
from utilities.numpy import NDim2
from utilities.numpy import NDim3

# annotations - dtype
DataArrayB: TypeAlias = Annotated[DataArray, DTypeB]
DataArrayDns: TypeAlias = Annotated[DataArray, DTypeDns]
DataArrayF: TypeAlias = Annotated[DataArray, DTypeF]
DataArrayI: TypeAlias = Annotated[DataArray, DTypeI]
DataArrayO: TypeAlias = Annotated[DataArray, DTypeO]

# annotations - ndim
DataArray0: TypeAlias = Annotated[DataArray, NDim0]
DataArray1: TypeAlias = Annotated[DataArray, NDim1]
DataArray2: TypeAlias = Annotated[DataArray, NDim2]
DataArray3: TypeAlias = Annotated[DataArray, NDim3]

# annotated; dtype & ndim
DataArrayB0: TypeAlias = Annotated[DataArray, DTypeB, NDim0]
DataArrayDns0: TypeAlias = Annotated[DataArray, DTypeDns, NDim0]
DataArrayF0: TypeAlias = Annotated[DataArray, DTypeF, NDim0]
DataArrayI0: TypeAlias = Annotated[DataArray, DTypeI, NDim0]
DataArrayO0: TypeAlias = Annotated[DataArray, DTypeO, NDim0]

DataArrayB1: TypeAlias = Annotated[DataArray, DTypeB, NDim1]
DataArrayDns1: TypeAlias = Annotated[DataArray, DTypeDns, NDim1]
DataArrayF1: TypeAlias = Annotated[DataArray, DTypeF, NDim1]
DataArrayI1: TypeAlias = Annotated[DataArray, DTypeI, NDim1]
DataArrayO1: TypeAlias = Annotated[DataArray, DTypeO, NDim1]

DataArrayB2: TypeAlias = Annotated[DataArray, DTypeB, NDim2]
DataArrayDns2: TypeAlias = Annotated[DataArray, DTypeDns, NDim2]
DataArrayF2: TypeAlias = Annotated[DataArray, DTypeF, NDim2]
DataArrayI2: TypeAlias = Annotated[DataArray, DTypeI, NDim2]
DataArrayO2: TypeAlias = Annotated[DataArray, DTypeO, NDim2]

DataArrayB3: TypeAlias = Annotated[DataArray, DTypeB, NDim3]
DataArrayDns3: TypeAlias = Annotated[DataArray, DTypeDns, NDim3]
DataArrayF3: TypeAlias = Annotated[DataArray, DTypeF, NDim3]
DataArrayI3: TypeAlias = Annotated[DataArray, DTypeI, NDim3]
DataArrayO3: TypeAlias = Annotated[DataArray, DTypeO, NDim3]



from __future__ import annotations

import datetime as dt
from collections.abc import Iterator
from contextlib import contextmanager
from typing import Any

from fpdf import FPDF
from fpdf.enums import XPos
from fpdf.enums import YPos
from typing_extensions import override

from utilities.datetime import local_timezone
from utilities.holoviews import save_plot
from utilities.tempfile import TemporaryDirectory


class _BasePDF(FPDF):
    """Base class for PDFs."""

    def add_fixed_width_text(self, text: str, /) -> None:
        """Add a block of fixed witth text."""
        self.set_font("Courier")
        _ = self.write(txt=text)
        self.ln()

    def add_plot(self, plot: Any, /) -> None:  # pragma: no cover
        with TemporaryDirectory() as temp:
            path = temp.joinpath("image.png")
            save_plot(plot, path)
            _ = self.image(path, w=self.epw)


@contextmanager
def yield_pdf(*, header: str | None = None) -> Iterator[_BasePDF]:
    """Yield a PDF."""

    class OutputPDF(_BasePDF):
        @override
        def header(self) -> None:
            if header is not None:
                self.set_font(family="Helvetica", style="B", size=15)
                _ = self.cell(w=80)
                _ = self.cell(
                    w=30,
                    h=10,
                    txt=header,
                    border=0,
                    align="C",
                    new_x=XPos.RIGHT,
                    new_y=YPos.TOP,
                )
                self.ln(20)

        @override
        def footer(self) -> None:
            self.set_y(-15)
            self.set_font(family="Helvetica", style="I", size=8)
            page_no, now = self.page_no(), dt.datetime.now(tz=local_timezone())
            txt = f"page {page_no}/{{nb}}; {now:%Y-%m-%d %H:%M:%S}"
            _ = self.cell(
                w=0,
                h=10,
                txt=txt,
                border=0,
                align="C",
                new_x=XPos.RIGHT,
                new_y=YPos.TOP,
            )

    pdf = OutputPDF(orientation="portrait", unit="mm", format="A4")
    pdf.alias_nb_pages()
    pdf.add_page()
    pdf.set_font("Helvetica", size=10)
    yield pdf



from __future__ import annotations

import datetime as dt
from collections.abc import Callable
from operator import eq
from operator import gt
from operator import lt
from re import search
from typing import Any

from hypothesis import HealthCheck
from hypothesis import assume
from hypothesis import given
from hypothesis import settings
from hypothesis.strategies import DataObject
from hypothesis.strategies import SearchStrategy
from hypothesis.strategies import data
from hypothesis.strategies import dates
from hypothesis.strategies import datetimes
from hypothesis.strategies import integers
from hypothesis.strategies import just
from hypothesis.strategies import sampled_from
from hypothesis.strategies import timedeltas
from hypothesis.strategies import times
from pytest import mark
from pytest import param
from pytest import raises

from utilities.datetime import EPOCH_UTC
from utilities.datetime import UTC
from utilities.datetime import CallYieldWeekdaysError
from utilities.datetime import IsWeekendError
from utilities.datetime import ParseDateError
from utilities.datetime import ParseDateTimeError
from utilities.datetime import ParseTimeError
from utilities.datetime import TimedeltaError
from utilities.datetime import add_weekdays
from utilities.datetime import date_to_datetime
from utilities.datetime import ensure_date
from utilities.datetime import ensure_datetime
from utilities.datetime import ensure_time
from utilities.datetime import ensure_timedelta
from utilities.datetime import is_weekday
from utilities.datetime import local_timezone
from utilities.datetime import maybe_sub_pct_y
from utilities.datetime import parse_date
from utilities.datetime import parse_datetime
from utilities.datetime import parse_time
from utilities.datetime import parse_timedelta
from utilities.datetime import round_to_next_weekday
from utilities.datetime import round_to_prev_weekday
from utilities.datetime import serialize_date
from utilities.datetime import serialize_datetime
from utilities.datetime import serialize_time
from utilities.datetime import serialize_timedelta
from utilities.datetime import yield_weekdays
from utilities.hypothesis import assume_does_not_raise
from utilities.hypothesis import text_clean
from utilities.platform import SYSTEM
from utilities.platform import System


class TestAddWeekdays:
    @given(date=dates(), n=integers(-10, 10))
    @mark.parametrize("predicate", [param(gt), param(lt)])
    def test_add(
        self, date: dt.date, n: int, predicate: Callable[[Any, Any], bool]
    ) -> None:
        _ = assume(predicate(n, 0))
        with assume_does_not_raise(OverflowError):
            result = add_weekdays(date, n=n)
        assert is_weekday(result)
        assert predicate(result, date)

    @given(date=dates())
    def test_zero(self, date: dt.date) -> None:
        _ = assume(is_weekday(date))
        result = add_weekdays(date, n=0)
        assert result == date

    @given(date=dates())
    @settings(suppress_health_check={HealthCheck.filter_too_much})
    def test_error(self, date: dt.date) -> None:
        _ = assume(not is_weekday(date))
        with raises(IsWeekendError):
            _ = add_weekdays(date, n=0)

    @given(date=dates(), n1=integers(-10, 10), n2=integers(-10, 10))
    def test_two(self, date: dt.date, n1: int, n2: int) -> None:
        with assume_does_not_raise(IsWeekendError, OverflowError):
            weekday1, weekday2 = (add_weekdays(date, n=n) for n in [n1, n2])
        result = weekday1 <= weekday2
        expected = n1 <= n2
        assert result is expected


class TestDateToDatetime:
    @given(date=dates())
    def test_main(self, date: dt.date) -> None:
        result = date_to_datetime(date).date()
        assert result == date


class TestEnsure:
    @given(data=data())
    @mark.parametrize(
        ("strategy", "func"),
        [
            param(dates(), ensure_date),
            param(datetimes(timezones=just(UTC)), ensure_datetime),
            param(times(), ensure_time),
            param(timedeltas(), ensure_timedelta),
        ],
    )
    def test_main(
        self,
        data: DataObject,
        strategy: SearchStrategy[Any],
        func: Callable[[Any], Any],
    ) -> None:
        value = data.draw(strategy)
        maybe_value = data.draw(sampled_from([value, str(value)]))
        result = func(maybe_value)
        assert result == value


class TestIsWeekday:
    @given(date=dates())
    def test_is_weekday(self, date: dt.date) -> None:
        result = is_weekday(date)
        name = date.strftime("%A")
        expected = name in {
            "Monday",
            "Tuesday",
            "Wednesday",
            "Thursday",
            "Friday",
        }
        assert result is expected


class TestLocalTimeZone:
    def test_main(self) -> None:
        tz = local_timezone()
        now = dt.datetime.now(tz=UTC)
        result = tz.tzname(now)
        expected = {"HKT", "JST", "UTC"}
        assert result in expected


class TestMaybeMaybeSubPctY:
    @given(text=text_clean())
    @mark.skipif(SYSTEM is not System.linux, reason="Linux only")
    def test_main(self, text: str) -> None:
        result = maybe_sub_pct_y(text)
        assert not search("%Y", result)


class TestParseDate:
    @given(date=dates())
    def test_str(self, date: dt.date) -> None:
        result = parse_date(str(date))
        assert result == date

    @given(date=dates())
    def test_isoformat(self, date: dt.date) -> None:
        result = parse_date(date.isoformat())
        assert result == date

    @given(date=dates())
    def test_yyyymmdd(self, date: dt.date) -> None:
        result = parse_date(date.strftime(maybe_sub_pct_y("%Y%m%d")))
        assert result == date

    def test_error(self) -> None:
        with raises(ParseDateError):
            _ = parse_date("error")


class TestParseDateTime:
    @given(datetime=datetimes(timezones=just(UTC)))
    def test_str(self, datetime: dt.datetime) -> None:
        result = parse_datetime(str(datetime))
        assert result == datetime

    @given(datetime=datetimes(timezones=just(UTC)))
    def test_isoformat(self, datetime: dt.datetime) -> None:
        result = parse_datetime(datetime.isoformat())
        assert result == datetime

    @given(
        datetime=datetimes(timezones=just(UTC)),
        fmt=sampled_from(["%Y%m%dT%H%M%S.%f%z", "%Y-%m-%d %H:%M:%S.%f%z"]).map(
            maybe_sub_pct_y
        ),
    )
    def test_yyyymmdd_hhmmss_fff_zzzz(
        self, datetime: dt.datetime, fmt: str
    ) -> None:
        result = parse_datetime(datetime.strftime(fmt))
        assert result == datetime

    @given(
        datetime=datetimes(timezones=just(UTC)),
        fmt=sampled_from(["%Y%m%dT%H%M%S.%f", "%Y-%m-%d %H:%M:%S.%f"]).map(
            maybe_sub_pct_y
        ),
    )
    def test_yyyymmdd_hhmmss_fff(self, datetime: dt.datetime, fmt: str) -> None:
        result = parse_datetime(datetime.strftime(fmt))
        assert result == datetime

    @given(
        datetime=datetimes(timezones=just(UTC)),
        fmt=sampled_from(
            ["%Y%m%dT%H%M%S", "%Y-%m-%d %H:%M:%S", "%Y-%m-%dT%H:%M:%S"]
        ).map(maybe_sub_pct_y),
    )
    def test_yyyymmdd_hhmmss(self, datetime: dt.datetime, fmt: str) -> None:
        datetime = datetime.replace(microsecond=0)
        result = parse_datetime(datetime.strftime(fmt))
        assert result == datetime

    @given(
        datetime=datetimes(timezones=just(UTC)),
        fmt=sampled_from(
            ["%Y%m%dT%H%M", "%Y-%m-%d %H:%M", "%Y-%m-%dT%H:%M"]
        ).map(maybe_sub_pct_y),
    )
    def test_yyyymmdd_hhmm(self, datetime: dt.datetime, fmt: str) -> None:
        datetime = datetime.replace(second=0, microsecond=0)
        result = parse_datetime(datetime.strftime(fmt))
        assert result == datetime

    @given(
        datetime=datetimes(timezones=just(UTC)),
        fmt=sampled_from(["%Y%m%dT%H", "%Y-%m-%d %H", "%Y-%m-%dT%H"]).map(
            maybe_sub_pct_y
        ),
    )
    def test_yyyymmdd_hh(self, datetime: dt.datetime, fmt: str) -> None:
        datetime = datetime.replace(minute=0, second=0, microsecond=0)
        result = parse_datetime(datetime.strftime(fmt))
        assert result == datetime

    @given(
        datetime=datetimes(timezones=just(UTC)),
        fmt=sampled_from(["%Y%m%d", "%Y-%m-%d"]).map(maybe_sub_pct_y),
    )
    def test_yyyymmdd(self, datetime: dt.datetime, fmt: str) -> None:
        datetime = datetime.replace(hour=0, minute=0, second=0, microsecond=0)
        result = parse_datetime(datetime.strftime(fmt))
        assert result == datetime

    def test_error(self) -> None:
        with raises(ParseDateTimeError):
            _ = parse_datetime("error")


class TestParseTime:
    @given(time=times())
    def test_str(self, time: dt.time) -> None:
        result = parse_time(str(time))
        assert result == time

    @given(time=times())
    def test_isoformat(self, time: dt.time) -> None:
        result = parse_time(time.isoformat())
        assert result == time

    @given(time=times(), fmt=sampled_from(["%H%M%S.%f", "%H:%M:%S.%f"]))
    def test_hhmmss_fff(self, time: dt.time, fmt: str) -> None:
        result = parse_time(time.strftime(fmt))
        assert result == time

    @given(time=times(), fmt=sampled_from(["%H%M%S", "%H:%M:%S"]))
    def test_hhmmss(self, time: dt.time, fmt: str) -> None:
        time = time.replace(microsecond=0)
        result = parse_time(time.strftime(fmt))
        assert result == time

    @given(time=times(), fmt=sampled_from(["%H%M", "%H:%M"]))
    def test_hhmm(self, time: dt.time, fmt: str) -> None:
        time = time.replace(second=0, microsecond=0)
        result = parse_time(time.strftime(fmt))
        assert result == time

    @given(time=times(), fmt=sampled_from(["%H", "%H"]))
    def test_hh(self, time: dt.time, fmt: str) -> None:
        time = time.replace(minute=0, second=0, microsecond=0)
        result = parse_time(time.strftime(fmt))
        assert result == time

    def test_error(self) -> None:
        with raises(ParseTimeError):
            _ = parse_time("error")


class TestParseTimedelta:
    @given(timedelta=timedeltas())
    def test_main(self, timedelta: dt.timedelta) -> None:
        result = parse_timedelta(str(timedelta))
        assert result == timedelta

    def test_error(self) -> None:
        with raises(TimedeltaError):
            _ = parse_timedelta("error")


class TestSerialize:
    @given(data=data())
    @mark.parametrize(
        ("strategy", "serialize", "parse"),
        [
            param(dates(), serialize_date, parse_date),
            param(
                datetimes(timezones=just(UTC)),
                serialize_datetime,
                parse_datetime,
            ),
            param(times(), serialize_time, parse_time),
            param(timedeltas(), str, parse_timedelta),
            param(timedeltas(), serialize_timedelta, parse_timedelta),
        ],
    )
    def test_main(
        self,
        data: DataObject,
        strategy: SearchStrategy[Any],
        serialize: Callable[[Any], Any],
        parse: Callable[[Any], Any],
    ) -> None:
        value = data.draw(strategy)
        result = parse(serialize(value))
        assert result == value

    @given(datetime=datetimes())
    def test_serialize_date(self, datetime: dt.datetime) -> None:
        result = parse_date(serialize_date(datetime))
        assert result == datetime.date()


class TestRoundToWeekday:
    @given(date=dates())
    @settings(suppress_health_check={HealthCheck.filter_too_much})
    @mark.parametrize(
        ("func", "predicate", "operator"),
        [
            param(round_to_next_weekday, True, eq),
            param(round_to_next_weekday, False, gt),
            param(round_to_prev_weekday, True, eq),
            param(round_to_prev_weekday, False, lt),
        ],
    )
    def test_main(
        self,
        date: dt.date,
        func: Callable[[dt.date], dt.date],
        predicate: bool,
        operator: Callable[[dt.date, dt.date], bool],
    ) -> None:
        _ = assume(is_weekday(date) is predicate)
        with assume_does_not_raise(OverflowError):
            result = func(date)
        assert operator(result, date)


class TestTimes:
    def test_main(self) -> None:
        assert isinstance(EPOCH_UTC, dt.datetime)
        assert EPOCH_UTC.tzinfo is UTC


class TestTimeZones:
    def test_main(self) -> None:
        assert isinstance(UTC, dt.tzinfo)


class TestYieldWeekdays:
    @given(start=dates(), days=integers(0, 365))
    def test_start_and_end(self, start: dt.date, days: int) -> None:
        with assume_does_not_raise(OverflowError):
            end = start + dt.timedelta(days=days)
        dates = list(yield_weekdays(start=start, end=end))
        assert all(start <= d <= end for d in dates)
        assert all(map(is_weekday, dates))

    @given(start=dates(), days=integers(0, 10))
    def test_start_and_days(self, start: dt.date, days: int) -> None:
        dates = list(yield_weekdays(start=start, days=days))
        assert len(dates) == days
        assert all(d >= start for d in dates)
        assert all(map(is_weekday, dates))

    @given(end=dates(), days=integers(0, 10))
    def test_end_and_days(self, end: dt.date, days: int) -> None:
        dates = list(yield_weekdays(end=end, days=days))
        assert len(dates) == days
        assert all(d <= end for d in dates)
        assert all(map(is_weekday, dates))

    def test_error(self) -> None:
        with raises(CallYieldWeekdaysError):
            _ = list(yield_weekdays())



from __future__ import annotations

from .numpy import DATE_MAX_AS_DATETIME64
from .numpy import DATE_MAX_AS_INT
from .numpy import DATE_MIN_AS_DATETIME64
from .numpy import DATE_MIN_AS_INT
from .numpy import DATETIME_MAX_AS_DATETIMETIME64
from .numpy import DATETIME_MAX_AS_INT
from .numpy import DATETIME_MIN_AS_DATETIMETIME64
from .numpy import DATETIME_MIN_AS_INT
from .numpy import DateOverflowError
from .numpy import Datetime64Kind
from .numpy import Datetime64Unit
from .numpy import DTypeB
from .numpy import DTypeDns
from .numpy import DTypeF
from .numpy import DTypeI
from .numpy import DTypeO
from .numpy import EmptyNumpyConcatenateError
from .numpy import InfElementsError
from .numpy import InvalidDTypeError
from .numpy import IsFinite
from .numpy import IsFiniteAndIntegral
from .numpy import IsFiniteAndIntegralOrNan
from .numpy import IsFiniteAndNegative
from .numpy import IsFiniteAndNegativeOrNan
from .numpy import IsFiniteAndNonNegative
from .numpy import IsFiniteAndNonNegativeOrNan
from .numpy import IsFiniteAndNonPositive
from .numpy import IsFiniteAndNonPositiveOrNan
from .numpy import IsFiniteAndNonZero
from .numpy import IsFiniteAndNonZeroOrNan
from .numpy import IsFiniteAndPositive
from .numpy import IsFiniteAndPositiveOrNan
from .numpy import IsFiniteOrNan
from .numpy import IsIntegral
from .numpy import IsIntegralOrNan
from .numpy import IsNegative
from .numpy import IsNegativeOrNan
from .numpy import IsNonNegative
from .numpy import IsNonNegativeOrNan
from .numpy import IsNonPositive
from .numpy import IsNonPositiveOrNan
from .numpy import IsNonZero
from .numpy import IsNonZeroOrNan
from .numpy import IsPositive
from .numpy import IsPositiveOrNan
from .numpy import IsZero
from .numpy import IsZeroOrFiniteAndNonMicro
from .numpy import IsZeroOrFiniteAndNonMicroOrNan
from .numpy import IsZeroOrNan
from .numpy import IsZeroOrNonMicro
from .numpy import IsZeroOrNonMicroOrNan
from .numpy import LossOfNanosecondsError
from .numpy import MultipleTrueElementsError
from .numpy import NanElementsError
from .numpy import NDArray0
from .numpy import NDArray1
from .numpy import NDArray2
from .numpy import NDArray3
from .numpy import NDArrayA
from .numpy import NDArrayB
from .numpy import NDArrayB0
from .numpy import NDArrayB1
from .numpy import NDArrayB2
from .numpy import NDArrayB3
from .numpy import NDArrayD
from .numpy import NDArrayD0
from .numpy import NDArrayD1
from .numpy import NDArrayD2
from .numpy import NDArrayD3
from .numpy import NDArrayDas
from .numpy import NDArrayDas0
from .numpy import NDArrayDas1
from .numpy import NDArrayDas2
from .numpy import NDArrayDas3
from .numpy import NDArrayDD
from .numpy import NDArrayDD0
from .numpy import NDArrayDD1
from .numpy import NDArrayDD2
from .numpy import NDArrayDD3
from .numpy import NDArrayDfs
from .numpy import NDArrayDfs0
from .numpy import NDArrayDfs1
from .numpy import NDArrayDfs2
from .numpy import NDArrayDfs3
from .numpy import NDArrayDh
from .numpy import NDArrayDh0
from .numpy import NDArrayDh1
from .numpy import NDArrayDh2
from .numpy import NDArrayDh3
from .numpy import NDArrayDM
from .numpy import NDArrayDm
from .numpy import NDArrayDM0
from .numpy import NDArrayDm0
from .numpy import NDArrayDM1
from .numpy import NDArrayDm1
from .numpy import NDArrayDM2
from .numpy import NDArrayDm2
from .numpy import NDArrayDM3
from .numpy import NDArrayDm3
from .numpy import NDArrayDms
from .numpy import NDArrayDms0
from .numpy import NDArrayDms1
from .numpy import NDArrayDms2
from .numpy import NDArrayDms3
from .numpy import NDArrayDns
from .numpy import NDArrayDns0
from .numpy import NDArrayDns1
from .numpy import NDArrayDns2
from .numpy import NDArrayDns3
from .numpy import NDArrayDps
from .numpy import NDArrayDps0
from .numpy import NDArrayDps1
from .numpy import NDArrayDps2
from .numpy import NDArrayDps3
from .numpy import NDArrayDs
from .numpy import NDArrayDs0
from .numpy import NDArrayDs1
from .numpy import NDArrayDs2
from .numpy import NDArrayDs3
from .numpy import NDArrayDus
from .numpy import NDArrayDus0
from .numpy import NDArrayDus1
from .numpy import NDArrayDus2
from .numpy import NDArrayDus3
from .numpy import NDArrayDW
from .numpy import NDArrayDW0
from .numpy import NDArrayDW1
from .numpy import NDArrayDW2
from .numpy import NDArrayDW3
from .numpy import NDArrayDY
from .numpy import NDArrayDY0
from .numpy import NDArrayDY1
from .numpy import NDArrayDY2
from .numpy import NDArrayDY3
from .numpy import NDArrayF
from .numpy import NDArrayF0
from .numpy import NDArrayF0Fin
from .numpy import NDArrayF0FinInt
from .numpy import NDArrayF0FinIntNan
from .numpy import NDArrayF0FinNan
from .numpy import NDArrayF0FinNeg
from .numpy import NDArrayF0FinNegNan
from .numpy import NDArrayF0FinNonNeg
from .numpy import NDArrayF0FinNonNegNan
from .numpy import NDArrayF0FinNonPos
from .numpy import NDArrayF0FinNonPosNan
from .numpy import NDArrayF0FinNonZr
from .numpy import NDArrayF0FinNonZrNan
from .numpy import NDArrayF0FinPos
from .numpy import NDArrayF0FinPosNan
from .numpy import NDArrayF0Int
from .numpy import NDArrayF0IntNan
from .numpy import NDArrayF0Neg
from .numpy import NDArrayF0NegNan
from .numpy import NDArrayF0NonNeg
from .numpy import NDArrayF0NonNegNan
from .numpy import NDArrayF0NonPos
from .numpy import NDArrayF0NonPosNan
from .numpy import NDArrayF0NonZr
from .numpy import NDArrayF0NonZrNan
from .numpy import NDArrayF0Pos
from .numpy import NDArrayF0PosNan
from .numpy import NDArrayF0Zr
from .numpy import NDArrayF0ZrFinNonMic
from .numpy import NDArrayF0ZrFinNonMicNan
from .numpy import NDArrayF0ZrNan
from .numpy import NDArrayF0ZrNonMic
from .numpy import NDArrayF0ZrNonMicNan
from .numpy import NDArrayF1
from .numpy import NDArrayF1Fin
from .numpy import NDArrayF1FinInt
from .numpy import NDArrayF1FinIntNan
from .numpy import NDArrayF1FinNan
from .numpy import NDArrayF1FinNeg
from .numpy import NDArrayF1FinNegNan
from .numpy import NDArrayF1FinNonNeg
from .numpy import NDArrayF1FinNonNegNan
from .numpy import NDArrayF1FinNonPos
from .numpy import NDArrayF1FinNonPosNan
from .numpy import NDArrayF1FinNonZr
from .numpy import NDArrayF1FinNonZrNan
from .numpy import NDArrayF1FinPos
from .numpy import NDArrayF1FinPosNan
from .numpy import NDArrayF1Int
from .numpy import NDArrayF1IntNan
from .numpy import NDArrayF1Neg
from .numpy import NDArrayF1NegNan
from .numpy import NDArrayF1NonNeg
from .numpy import NDArrayF1NonNegNan
from .numpy import NDArrayF1NonPos
from .numpy import NDArrayF1NonPosNan
from .numpy import NDArrayF1NonZr
from .numpy import NDArrayF1NonZrNan
from .numpy import NDArrayF1Pos
from .numpy import NDArrayF1PosNan
from .numpy import NDArrayF1Zr
from .numpy import NDArrayF1ZrFinNonMic
from .numpy import NDArrayF1ZrFinNonMicNan
from .numpy import NDArrayF1ZrNan
from .numpy import NDArrayF1ZrNonMic
from .numpy import NDArrayF1ZrNonMicNan
from .numpy import NDArrayF2
from .numpy import NDArrayF2Fin
from .numpy import NDArrayF2FinInt
from .numpy import NDArrayF2FinIntNan
from .numpy import NDArrayF2FinNan
from .numpy import NDArrayF2FinNeg
from .numpy import NDArrayF2FinNegNan
from .numpy import NDArrayF2FinNonNeg
from .numpy import NDArrayF2FinNonNegNan
from .numpy import NDArrayF2FinNonPos
from .numpy import NDArrayF2FinNonPosNan
from .numpy import NDArrayF2FinNonZr
from .numpy import NDArrayF2FinNonZrNan
from .numpy import NDArrayF2FinPos
from .numpy import NDArrayF2FinPosNan
from .numpy import NDArrayF2Int
from .numpy import NDArrayF2IntNan
from .numpy import NDArrayF2Neg
from .numpy import NDArrayF2NegNan
from .numpy import NDArrayF2NonNeg
from .numpy import NDArrayF2NonNegNan
from .numpy import NDArrayF2NonPos
from .numpy import NDArrayF2NonPosNan
from .numpy import NDArrayF2NonZr
from .numpy import NDArrayF2NonZrNan
from .numpy import NDArrayF2Pos
from .numpy import NDArrayF2PosNan
from .numpy import NDArrayF2Zr
from .numpy import NDArrayF2ZrFinNonMic
from .numpy import NDArrayF2ZrFinNonMicNan
from .numpy import NDArrayF2ZrNan
from .numpy import NDArrayF2ZrNonMic
from .numpy import NDArrayF2ZrNonMicNan
from .numpy import NDArrayF3
from .numpy import NDArrayF3Fin
from .numpy import NDArrayF3FinInt
from .numpy import NDArrayF3FinIntNan
from .numpy import NDArrayF3FinNan
from .numpy import NDArrayF3FinNeg
from .numpy import NDArrayF3FinNegNan
from .numpy import NDArrayF3FinNonNeg
from .numpy import NDArrayF3FinNonNegNan
from .numpy import NDArrayF3FinNonPos
from .numpy import NDArrayF3FinNonPosNan
from .numpy import NDArrayF3FinNonZr
from .numpy import NDArrayF3FinNonZrNan
from .numpy import NDArrayF3FinPos
from .numpy import NDArrayF3FinPosNan
from .numpy import NDArrayF3Int
from .numpy import NDArrayF3IntNan
from .numpy import NDArrayF3Neg
from .numpy import NDArrayF3NegNan
from .numpy import NDArrayF3NonNeg
from .numpy import NDArrayF3NonNegNan
from .numpy import NDArrayF3NonPos
from .numpy import NDArrayF3NonPosNan
from .numpy import NDArrayF3NonZr
from .numpy import NDArrayF3NonZrNan
from .numpy import NDArrayF3Pos
from .numpy import NDArrayF3PosNan
from .numpy import NDArrayF3Zr
from .numpy import NDArrayF3ZrFinNonMic
from .numpy import NDArrayF3ZrFinNonMicNan
from .numpy import NDArrayF3ZrNan
from .numpy import NDArrayF3ZrNonMic
from .numpy import NDArrayF3ZrNonMicNan
from .numpy import NDArrayFFin
from .numpy import NDArrayFFinInt
from .numpy import NDArrayFFinIntNan
from .numpy import NDArrayFFinNan
from .numpy import NDArrayFFinNeg
from .numpy import NDArrayFFinNegNan
from .numpy import NDArrayFFinNonNeg
from .numpy import NDArrayFFinNonNegNan
from .numpy import NDArrayFFinNonPos
from .numpy import NDArrayFFinNonPosNan
from .numpy import NDArrayFFinNonZr
from .numpy import NDArrayFFinNonZrNan
from .numpy import NDArrayFFinPos
from .numpy import NDArrayFFinPosNan
from .numpy import NDArrayFInt
from .numpy import NDArrayFIntNan
from .numpy import NDArrayFNeg
from .numpy import NDArrayFNegNan
from .numpy import NDArrayFNonNeg
from .numpy import NDArrayFNonNegNan
from .numpy import NDArrayFNonPos
from .numpy import NDArrayFNonPosNan
from .numpy import NDArrayFNonZr
from .numpy import NDArrayFNonZrNan
from .numpy import NDArrayFPos
from .numpy import NDArrayFPosNan
from .numpy import NDArrayFZr
from .numpy import NDArrayFZrFinNonMic
from .numpy import NDArrayFZrFinNonMicNan
from .numpy import NDArrayFZrNan
from .numpy import NDArrayFZrNonMic
from .numpy import NDArrayFZrNonMicNan
from .numpy import NDArrayI
from .numpy import NDArrayI0
from .numpy import NDArrayI0Neg
from .numpy import NDArrayI0NonNeg
from .numpy import NDArrayI0NonPos
from .numpy import NDArrayI0NonZr
from .numpy import NDArrayI0Pos
from .numpy import NDArrayI0Zr
from .numpy import NDArrayI1
from .numpy import NDArrayI1Neg
from .numpy import NDArrayI1NonNeg
from .numpy import NDArrayI1NonPos
from .numpy import NDArrayI1NonZr
from .numpy import NDArrayI1Pos
from .numpy import NDArrayI1Zr
from .numpy import NDArrayI2
from .numpy import NDArrayI2Neg
from .numpy import NDArrayI2NonNeg
from .numpy import NDArrayI2NonPos
from .numpy import NDArrayI2NonZr
from .numpy import NDArrayI2Pos
from .numpy import NDArrayI2Zr
from .numpy import NDArrayI3
from .numpy import NDArrayI3Neg
from .numpy import NDArrayI3NonNeg
from .numpy import NDArrayI3NonPos
from .numpy import NDArrayI3NonZr
from .numpy import NDArrayI3Pos
from .numpy import NDArrayI3Zr
from .numpy import NDArrayINeg
from .numpy import NDArrayINonNeg
from .numpy import NDArrayINonPos
from .numpy import NDArrayINonZr
from .numpy import NDArrayIPos
from .numpy import NDArrayIZr
from .numpy import NDArrayO
from .numpy import NDArrayO0
from .numpy import NDArrayO1
from .numpy import NDArrayO2
from .numpy import NDArrayO3
from .numpy import NDim0
from .numpy import NDim1
from .numpy import NDim2
from .numpy import NDim3
from .numpy import NonIntegralElementsError
from .numpy import NoTrueElementsError
from .numpy import ZeroShiftError
from .numpy import array_indexer
from .numpy import as_int
from .numpy import date_to_datetime64
from .numpy import datetime64_dtype_to_unit
from .numpy import datetime64_to_date
from .numpy import datetime64_to_datetime
from .numpy import datetime64_to_int
from .numpy import datetime64_unit_to_dtype
from .numpy import datetime64_unit_to_kind
from .numpy import datetime64as
from .numpy import datetime64D
from .numpy import datetime64fs
from .numpy import datetime64h
from .numpy import datetime64M
from .numpy import datetime64m
from .numpy import datetime64ms
from .numpy import datetime64ns
from .numpy import datetime64ps
from .numpy import datetime64s
from .numpy import datetime64us
from .numpy import datetime64W
from .numpy import datetime64Y
from .numpy import datetime_to_datetime64
from .numpy import discretize
from .numpy import ffill_non_nan_slices
from .numpy import fillna
from .numpy import flatn0
from .numpy import get_fill_value
from .numpy import has_dtype
from .numpy import is_at_least
from .numpy import is_at_least_or_nan
from .numpy import is_at_most
from .numpy import is_at_most_or_nan
from .numpy import is_between
from .numpy import is_between_or_nan
from .numpy import is_empty
from .numpy import is_finite_and_integral
from .numpy import is_finite_and_integral_or_nan
from .numpy import is_finite_and_negative
from .numpy import is_finite_and_negative_or_nan
from .numpy import is_finite_and_non_negative
from .numpy import is_finite_and_non_negative_or_nan
from .numpy import is_finite_and_non_positive
from .numpy import is_finite_and_non_positive_or_nan
from .numpy import is_finite_and_non_zero
from .numpy import is_finite_and_non_zero_or_nan
from .numpy import is_finite_and_positive
from .numpy import is_finite_and_positive_or_nan
from .numpy import is_finite_or_nan
from .numpy import is_greater_than
from .numpy import is_greater_than_or_nan
from .numpy import is_integral
from .numpy import is_integral_or_nan
from .numpy import is_less_than
from .numpy import is_less_than_or_nan
from .numpy import is_negative
from .numpy import is_negative_or_nan
from .numpy import is_non_empty
from .numpy import is_non_negative
from .numpy import is_non_negative_or_nan
from .numpy import is_non_positive
from .numpy import is_non_positive_or_nan
from .numpy import is_non_singular
from .numpy import is_non_zero
from .numpy import is_non_zero_or_nan
from .numpy import is_positive
from .numpy import is_positive_or_nan
from .numpy import is_positive_semidefinite
from .numpy import is_symmetric
from .numpy import is_zero
from .numpy import is_zero_or_finite_and_non_micro
from .numpy import is_zero_or_finite_and_non_micro_or_nan
from .numpy import is_zero_or_nan
from .numpy import is_zero_or_non_micro
from .numpy import is_zero_or_non_micro_or_nan
from .numpy import maximum
from .numpy import minimum
from .numpy import redirect_to_empty_numpy_concatenate_error
from .numpy import shift
from .numpy import shift_bool
from .numpy import year

__all__ = [
    "array_indexer",
    "as_int",
    "DATE_MAX_AS_DATETIME64",
    "DATE_MAX_AS_INT",
    "DATE_MIN_AS_DATETIME64",
    "DATE_MIN_AS_INT",
    "date_to_datetime64",
    "DateOverflowError",
    "DATETIME_MAX_AS_DATETIMETIME64",
    "DATETIME_MAX_AS_INT",
    "DATETIME_MIN_AS_DATETIMETIME64",
    "DATETIME_MIN_AS_INT",
    "datetime_to_datetime64",
    "datetime64_dtype_to_unit",
    "datetime64_to_date",
    "datetime64_to_datetime",
    "datetime64_to_int",
    "datetime64_unit_to_dtype",
    "datetime64_unit_to_kind",
    "datetime64as",
    "datetime64D",
    "datetime64fs",
    "datetime64h",
    "Datetime64Kind",
    "datetime64m",
    "datetime64M",
    "datetime64ms",
    "datetime64ns",
    "datetime64ps",
    "datetime64s",
    "Datetime64Unit",
    "datetime64us",
    "datetime64W",
    "datetime64Y",
    "discretize",
    "DTypeB",
    "DTypeDns",
    "DTypeF",
    "DTypeI",
    "DTypeO",
    "EmptyNumpyConcatenateError",
    "ffill_non_nan_slices",
    "fillna",
    "flatn0",
    "get_fill_value",
    "has_dtype",
    "InfElementsError",
    "InvalidDTypeError",
    "is_at_least_or_nan",
    "is_at_least",
    "is_at_most_or_nan",
    "is_at_most",
    "is_between_or_nan",
    "is_between",
    "is_empty",
    "is_finite_and_integral_or_nan",
    "is_finite_and_integral",
    "is_finite_and_negative_or_nan",
    "is_finite_and_negative",
    "is_finite_and_non_negative_or_nan",
    "is_finite_and_non_negative",
    "is_finite_and_non_positive_or_nan",
    "is_finite_and_non_positive",
    "is_finite_and_non_zero_or_nan",
    "is_finite_and_non_zero",
    "is_finite_and_positive_or_nan",
    "is_finite_and_positive",
    "is_finite_or_nan",
    "is_greater_than_or_nan",
    "is_greater_than",
    "is_integral_or_nan",
    "is_integral",
    "is_less_than_or_nan",
    "is_less_than",
    "is_negative_or_nan",
    "is_negative",
    "is_non_empty",
    "is_non_negative_or_nan",
    "is_non_negative",
    "is_non_positive_or_nan",
    "is_non_positive",
    "is_non_singular",
    "is_non_zero_or_nan",
    "is_non_zero",
    "is_positive_or_nan",
    "is_positive_semidefinite",
    "is_positive",
    "is_symmetric",
    "is_zero_or_finite_and_non_micro_or_nan",
    "is_zero_or_finite_and_non_micro",
    "is_zero_or_nan",
    "is_zero_or_non_micro_or_nan",
    "is_zero_or_non_micro",
    "is_zero",
    "IsFinite",
    "IsFiniteAndIntegral",
    "IsFiniteAndIntegralOrNan",
    "IsFiniteAndNegative",
    "IsFiniteAndNegativeOrNan",
    "IsFiniteAndNonNegative",
    "IsFiniteAndNonNegativeOrNan",
    "IsFiniteAndNonPositive",
    "IsFiniteAndNonPositiveOrNan",
    "IsFiniteAndNonZero",
    "IsFiniteAndNonZeroOrNan",
    "IsFiniteAndPositive",
    "IsFiniteAndPositiveOrNan",
    "IsFiniteOrNan",
    "IsIntegral",
    "IsIntegralOrNan",
    "IsNegative",
    "IsNegativeOrNan",
    "IsNonNegative",
    "IsNonNegativeOrNan",
    "IsNonPositive",
    "IsNonPositiveOrNan",
    "IsNonZero",
    "IsNonZeroOrNan",
    "IsPositive",
    "IsPositiveOrNan",
    "IsZero",
    "IsZeroOrFiniteAndNonMicro",
    "IsZeroOrFiniteAndNonMicroOrNan",
    "IsZeroOrNan",
    "IsZeroOrNonMicro",
    "IsZeroOrNonMicroOrNan",
    "LossOfNanosecondsError",
    "maximum",
    "minimum",
    "MultipleTrueElementsError",
    "NanElementsError",
    "NDArray0",
    "NDArray1",
    "NDArray2",
    "NDArray3",
    "NDArrayA",
    "NDArrayB",
    "NDArrayB0",
    "NDArrayB1",
    "NDArrayB2",
    "NDArrayB3",
    "NDArrayD",
    "NDArrayD0",
    "NDArrayD1",
    "NDArrayD2",
    "NDArrayD3",
    "NDArrayDas",
    "NDArrayDas0",
    "NDArrayDas1",
    "NDArrayDas2",
    "NDArrayDas3",
    "NDArrayDD",
    "NDArrayDD0",
    "NDArrayDD1",
    "NDArrayDD2",
    "NDArrayDD3",
    "NDArrayDfs",
    "NDArrayDfs0",
    "NDArrayDfs1",
    "NDArrayDfs2",
    "NDArrayDfs3",
    "NDArrayDh",
    "NDArrayDh0",
    "NDArrayDh1",
    "NDArrayDh2",
    "NDArrayDh3",
    "NDArrayDm",
    "NDArrayDM",
    "NDArrayDm0",
    "NDArrayDM0",
    "NDArrayDm1",
    "NDArrayDM1",
    "NDArrayDm2",
    "NDArrayDM2",
    "NDArrayDm3",
    "NDArrayDM3",
    "NDArrayDms",
    "NDArrayDms0",
    "NDArrayDms1",
    "NDArrayDms2",
    "NDArrayDms3",
    "NDArrayDns",
    "NDArrayDns0",
    "NDArrayDns1",
    "NDArrayDns2",
    "NDArrayDns3",
    "NDArrayDps",
    "NDArrayDps0",
    "NDArrayDps1",
    "NDArrayDps2",
    "NDArrayDps3",
    "NDArrayDs",
    "NDArrayDs0",
    "NDArrayDs1",
    "NDArrayDs2",
    "NDArrayDs3",
    "NDArrayDus",
    "NDArrayDus0",
    "NDArrayDus1",
    "NDArrayDus2",
    "NDArrayDus3",
    "NDArrayDW",
    "NDArrayDW0",
    "NDArrayDW1",
    "NDArrayDW2",
    "NDArrayDW3",
    "NDArrayDY",
    "NDArrayDY0",
    "NDArrayDY1",
    "NDArrayDY2",
    "NDArrayDY3",
    "NDArrayF",
    "NDArrayF0",
    "NDArrayF0Fin",
    "NDArrayF0FinInt",
    "NDArrayF0FinIntNan",
    "NDArrayF0FinNan",
    "NDArrayF0FinNeg",
    "NDArrayF0FinNegNan",
    "NDArrayF0FinNonNeg",
    "NDArrayF0FinNonNegNan",
    "NDArrayF0FinNonPos",
    "NDArrayF0FinNonPosNan",
    "NDArrayF0FinNonZr",
    "NDArrayF0FinNonZrNan",
    "NDArrayF0FinPos",
    "NDArrayF0FinPosNan",
    "NDArrayF0Int",
    "NDArrayF0IntNan",
    "NDArrayF0Neg",
    "NDArrayF0NegNan",
    "NDArrayF0NonNeg",
    "NDArrayF0NonNegNan",
    "NDArrayF0NonPos",
    "NDArrayF0NonPosNan",
    "NDArrayF0NonZr",
    "NDArrayF0NonZrNan",
    "NDArrayF0Pos",
    "NDArrayF0PosNan",
    "NDArrayF0Zr",
    "NDArrayF0ZrFinNonMic",
    "NDArrayF0ZrFinNonMicNan",
    "NDArrayF0ZrNan",
    "NDArrayF0ZrNonMic",
    "NDArrayF0ZrNonMicNan",
    "NDArrayF1",
    "NDArrayF1Fin",
    "NDArrayF1FinInt",
    "NDArrayF1FinIntNan",
    "NDArrayF1FinNan",
    "NDArrayF1FinNeg",
    "NDArrayF1FinNegNan",
    "NDArrayF1FinNonNeg",
    "NDArrayF1FinNonNegNan",
    "NDArrayF1FinNonPos",
    "NDArrayF1FinNonPosNan",
    "NDArrayF1FinNonZr",
    "NDArrayF1FinNonZrNan",
    "NDArrayF1FinPos",
    "NDArrayF1FinPosNan",
    "NDArrayF1Int",
    "NDArrayF1IntNan",
    "NDArrayF1Neg",
    "NDArrayF1NegNan",
    "NDArrayF1NonNeg",
    "NDArrayF1NonNegNan",
    "NDArrayF1NonPos",
    "NDArrayF1NonPosNan",
    "NDArrayF1NonZr",
    "NDArrayF1NonZrNan",
    "NDArrayF1Pos",
    "NDArrayF1PosNan",
    "NDArrayF1Zr",
    "NDArrayF1ZrFinNonMic",
    "NDArrayF1ZrFinNonMicNan",
    "NDArrayF1ZrNan",
    "NDArrayF1ZrNonMic",
    "NDArrayF1ZrNonMicNan",
    "NDArrayF2",
    "NDArrayF2Fin",
    "NDArrayF2FinInt",
    "NDArrayF2FinIntNan",
    "NDArrayF2FinNan",
    "NDArrayF2FinNeg",
    "NDArrayF2FinNegNan",
    "NDArrayF2FinNonNeg",
    "NDArrayF2FinNonNegNan",
    "NDArrayF2FinNonPos",
    "NDArrayF2FinNonPosNan",
    "NDArrayF2FinNonZr",
    "NDArrayF2FinNonZrNan",
    "NDArrayF2FinPos",
    "NDArrayF2FinPosNan",
    "NDArrayF2Int",
    "NDArrayF2IntNan",
    "NDArrayF2Neg",
    "NDArrayF2NegNan",
    "NDArrayF2NonNeg",
    "NDArrayF2NonNegNan",
    "NDArrayF2NonPos",
    "NDArrayF2NonPosNan",
    "NDArrayF2NonZr",
    "NDArrayF2NonZrNan",
    "NDArrayF2Pos",
    "NDArrayF2PosNan",
    "NDArrayF2Zr",
    "NDArrayF2ZrFinNonMic",
    "NDArrayF2ZrFinNonMicNan",
    "NDArrayF2ZrNan",
    "NDArrayF2ZrNonMic",
    "NDArrayF2ZrNonMicNan",
    "NDArrayF3",
    "NDArrayF3Fin",
    "NDArrayF3FinInt",
    "NDArrayF3FinIntNan",
    "NDArrayF3FinNan",
    "NDArrayF3FinNeg",
    "NDArrayF3FinNegNan",
    "NDArrayF3FinNonNeg",
    "NDArrayF3FinNonNegNan",
    "NDArrayF3FinNonPos",
    "NDArrayF3FinNonPosNan",
    "NDArrayF3FinNonZr",
    "NDArrayF3FinNonZrNan",
    "NDArrayF3FinPos",
    "NDArrayF3FinPosNan",
    "NDArrayF3Int",
    "NDArrayF3IntNan",
    "NDArrayF3Neg",
    "NDArrayF3NegNan",
    "NDArrayF3NonNeg",
    "NDArrayF3NonNegNan",
    "NDArrayF3NonPos",
    "NDArrayF3NonPosNan",
    "NDArrayF3NonZr",
    "NDArrayF3NonZrNan",
    "NDArrayF3Pos",
    "NDArrayF3PosNan",
    "NDArrayF3Zr",
    "NDArrayF3ZrFinNonMic",
    "NDArrayF3ZrFinNonMicNan",
    "NDArrayF3ZrNan",
    "NDArrayF3ZrNonMic",
    "NDArrayF3ZrNonMicNan",
    "NDArrayFFin",
    "NDArrayFFinInt",
    "NDArrayFFinIntNan",
    "NDArrayFFinNan",
    "NDArrayFFinNeg",
    "NDArrayFFinNegNan",
    "NDArrayFFinNonNeg",
    "NDArrayFFinNonNegNan",
    "NDArrayFFinNonPos",
    "NDArrayFFinNonPosNan",
    "NDArrayFFinNonZr",
    "NDArrayFFinNonZrNan",
    "NDArrayFFinPos",
    "NDArrayFFinPosNan",
    "NDArrayFInt",
    "NDArrayFIntNan",
    "NDArrayFNeg",
    "NDArrayFNegNan",
    "NDArrayFNonNeg",
    "NDArrayFNonNegNan",
    "NDArrayFNonPos",
    "NDArrayFNonPosNan",
    "NDArrayFNonZr",
    "NDArrayFNonZrNan",
    "NDArrayFPos",
    "NDArrayFPosNan",
    "NDArrayFZr",
    "NDArrayFZrFinNonMic",
    "NDArrayFZrFinNonMicNan",
    "NDArrayFZrNan",
    "NDArrayFZrNonMic",
    "NDArrayFZrNonMicNan",
    "NDArrayI",
    "NDArrayI0",
    "NDArrayI0Neg",
    "NDArrayI0NonNeg",
    "NDArrayI0NonPos",
    "NDArrayI0NonZr",
    "NDArrayI0Pos",
    "NDArrayI0Zr",
    "NDArrayI1",
    "NDArrayI1Neg",
    "NDArrayI1NonNeg",
    "NDArrayI1NonPos",
    "NDArrayI1NonZr",
    "NDArrayI1Pos",
    "NDArrayI1Zr",
    "NDArrayI2",
    "NDArrayI2Neg",
    "NDArrayI2NonNeg",
    "NDArrayI2NonPos",
    "NDArrayI2NonZr",
    "NDArrayI2Pos",
    "NDArrayI2Zr",
    "NDArrayI3",
    "NDArrayI3Neg",
    "NDArrayI3NonNeg",
    "NDArrayI3NonPos",
    "NDArrayI3NonZr",
    "NDArrayI3Pos",
    "NDArrayI3Zr",
    "NDArrayINeg",
    "NDArrayINonNeg",
    "NDArrayINonPos",
    "NDArrayINonZr",
    "NDArrayIPos",
    "NDArrayIZr",
    "NDArrayO",
    "NDArrayO0",
    "NDArrayO1",
    "NDArrayO2",
    "NDArrayO3",
    "NDim0",
    "NDim1",
    "NDim2",
    "NDim3",
    "NonIntegralElementsError",
    "NoTrueElementsError",
    "redirect_to_empty_numpy_concatenate_error",
    "shift_bool",
    "shift",
    "year",
    "ZeroShiftError",
]

try:
    from .bottleneck import ZeroPercentageChangeSpanError
    from .bottleneck import ffill
    from .bottleneck import pct_change
except ModuleNotFoundError:  # pragma: no cover
    pass
else:
    __all__ += ["ffill", "pct_change", "ZeroPercentageChangeSpanError"]


try:
    from .numbagg import ewma
    from .numbagg import exp_moving_sum
except ModuleNotFoundError:  # pragma: no cover
    pass
else:
    __all__ += ["ewma", "exp_moving_sum"]



from __future__ import annotations

from typing import Any

from beartype.door import die_if_unbearable
from numpy import empty
from numpy import zeros
from pytest import mark
from pytest import param
from xarray import DataArray

from utilities.numpy import datetime64ns
from utilities.xarray import DataArray0
from utilities.xarray import DataArray1
from utilities.xarray import DataArray2
from utilities.xarray import DataArray3
from utilities.xarray import DataArrayB
from utilities.xarray import DataArrayB0
from utilities.xarray import DataArrayB1
from utilities.xarray import DataArrayB2
from utilities.xarray import DataArrayB3
from utilities.xarray import DataArrayDns
from utilities.xarray import DataArrayDns0
from utilities.xarray import DataArrayDns1
from utilities.xarray import DataArrayDns2
from utilities.xarray import DataArrayDns3
from utilities.xarray import DataArrayF
from utilities.xarray import DataArrayF0
from utilities.xarray import DataArrayF1
from utilities.xarray import DataArrayF2
from utilities.xarray import DataArrayF3
from utilities.xarray import DataArrayI
from utilities.xarray import DataArrayI0
from utilities.xarray import DataArrayI1
from utilities.xarray import DataArrayI2
from utilities.xarray import DataArrayI3
from utilities.xarray import DataArrayO
from utilities.xarray import DataArrayO0
from utilities.xarray import DataArrayO1
from utilities.xarray import DataArrayO2
from utilities.xarray import DataArrayO3


class TestHints:
    @mark.parametrize(
        ("dtype", "hint"),
        [
            param(bool, DataArrayB),
            param(datetime64ns, DataArrayDns),
            param(float, DataArrayF),
            param(int, DataArrayI),
            param(object, DataArrayO),
        ],
    )
    def test_dtype(self, dtype: Any, hint: Any) -> None:
        arr = DataArray(empty(0, dtype=dtype))
        die_if_unbearable(arr, hint)

    @mark.parametrize(
        ("ndim", "hint"),
        [
            param(0, DataArray0),
            param(1, DataArray1),
            param(2, DataArray2),
            param(3, DataArray3),
        ],
    )
    def test_ndim(self, ndim: int, hint: Any) -> None:
        arr = DataArray(empty(zeros(ndim, dtype=int), dtype=float))
        die_if_unbearable(arr, hint)

    @mark.parametrize(
        ("dtype", "ndim", "hint"),
        [
            # ndim 0
            param(bool, 0, DataArrayB0),
            param(datetime64ns, 0, DataArrayDns0),
            param(float, 0, DataArrayF0),
            param(int, 0, DataArrayI0),
            param(object, 0, DataArrayO0),
            # ndim 1
            param(bool, 1, DataArrayB1),
            param(datetime64ns, 1, DataArrayDns1),
            param(float, 1, DataArrayF1),
            param(int, 1, DataArrayI1),
            param(object, 1, DataArrayO1),
            # ndim 2
            param(bool, 2, DataArrayB2),
            param(datetime64ns, 2, DataArrayDns2),
            param(float, 2, DataArrayF2),
            param(int, 2, DataArrayI2),
            param(object, 2, DataArrayO2),
            # ndim 3
            param(bool, 3, DataArrayB3),
            param(datetime64ns, 3, DataArrayDns3),
            param(float, 3, DataArrayF3),
            param(int, 3, DataArrayI3),
            param(object, 3, DataArrayO3),
        ],
    )
    def test_compound(self, dtype: Any, ndim: int, hint: Any) -> None:
        arr = DataArray(empty(zeros(ndim, dtype=int), dtype=dtype))
        die_if_unbearable(arr, hint)



from __future__ import annotations

import datetime as dt
from collections.abc import Iterator
from contextlib import contextmanager
from pathlib import Path
from zoneinfo import ZoneInfo

from pyinstrument.profiler import Profiler

from utilities.atomicwrites import writer
from utilities.pathlib import PathLike


@contextmanager
def profile(*, path: PathLike = Path.cwd()) -> Iterator[None]:
    """Profile the contents of a block."""
    with Profiler() as profiler:
        yield
    now = dt.datetime.now(tz=ZoneInfo("UTC"))
    filename = Path(path, f"profile__{now:%Y%m%dT%H%M%S}.html")
    with writer(filename) as temp, temp.open(mode="w") as fh:
        _ = fh.write(profiler.output_html())



from __future__ import annotations

import datetime as dt
from typing import Any
from typing import Literal

import numpy as np
from hypothesis import assume
from hypothesis import given
from hypothesis.errors import InvalidArgument
from hypothesis.extra.numpy import array_shapes
from hypothesis.strategies import DataObject
from hypothesis.strategies import booleans
from hypothesis.strategies import data
from hypothesis.strategies import dates
from hypothesis.strategies import floats
from hypothesis.strategies import integers
from hypothesis.strategies import just
from hypothesis.strategies import none
from numpy import datetime64
from numpy import iinfo
from numpy import inf
from numpy import int32
from numpy import int64
from numpy import isfinite
from numpy import isinf
from numpy import isnan
from numpy import isnat
from numpy import ravel
from numpy import rint
from numpy import uint32
from numpy import uint64
from numpy import zeros
from numpy.testing import assert_equal
from pytest import raises

from utilities.hypothesis import Shape
from utilities.hypothesis import assume_does_not_raise
from utilities.hypothesis import bool_arrays
from utilities.hypothesis import concatenated_arrays
from utilities.hypothesis import datetime64_arrays
from utilities.hypothesis import datetime64_dtypes
from utilities.hypothesis import datetime64_indexes
from utilities.hypothesis import datetime64_kinds
from utilities.hypothesis import datetime64_units
from utilities.hypothesis import datetime64D_indexes
from utilities.hypothesis import datetime64s
from utilities.hypothesis import datetime64us_indexes
from utilities.hypothesis import datetimes_utc
from utilities.hypothesis import float_arrays
from utilities.hypothesis import int32s
from utilities.hypothesis import int64s
from utilities.hypothesis import int_arrays
from utilities.hypothesis import str_arrays
from utilities.hypothesis import uint32s
from utilities.hypothesis import uint64s
from utilities.numpy import Datetime64Kind
from utilities.numpy import Datetime64Unit
from utilities.numpy import datetime64_dtype_to_unit
from utilities.numpy import datetime64_to_date
from utilities.numpy import datetime64_to_datetime
from utilities.numpy import datetime64_to_int
from utilities.numpy import datetime64_unit_to_kind
from utilities.numpy import datetime64D
from utilities.numpy import datetime64us


class TestBoolArrays:
    @given(data=data(), shape=array_shapes())
    def test_main(self, data: DataObject, shape: Shape) -> None:
        array = data.draw(bool_arrays(shape=shape))
        assert array.dtype == bool
        assert array.shape == shape


class TestConcatenatedArrays:
    @given(data=data(), m=integers(0, 10), n=integers(0, 10))
    def test_1d(self, data: DataObject, m: int, n: int) -> None:
        arrays = just(zeros(n, dtype=float))
        array = data.draw(concatenated_arrays(arrays, m, n))
        assert array.shape == (m, n)

    @given(data=data(), m=integers(0, 10), n=integers(0, 10), p=integers(0, 10))
    def test_2d(self, data: DataObject, m: int, n: int, p: int) -> None:
        arrays = just(zeros((n, p), dtype=float))
        array = data.draw(concatenated_arrays(arrays, m, (n, p)))
        assert array.shape == (m, n, p)


class TestDatetime64Arrays:
    @given(
        data=data(),
        shape=array_shapes(),
        unit=datetime64_units() | none(),
        valid_dates=booleans(),
        valid_datetimes=booleans(),
        unique=booleans(),
    )
    def test_main(
        self,
        data: DataObject,
        shape: Shape,
        unit: Datetime64Unit | None,
        valid_dates: bool,
        valid_datetimes: bool,
        unique: bool,
    ) -> None:
        with assume_does_not_raise(InvalidArgument):
            array = data.draw(
                datetime64_arrays(
                    shape=shape,
                    unit=unit,
                    valid_dates=valid_dates,
                    valid_datetimes=valid_datetimes,
                    unique=unique,
                )
            )
        assert array.shape == shape
        if unit is not None:
            assert datetime64_dtype_to_unit(array.dtype) == unit
        if valid_dates:
            _ = [datetime64_to_date(d) for d in ravel(array)]
        if valid_datetimes:
            _ = [datetime64_to_datetime(d) for d in ravel(array)]
        if unique:
            assert len(ravel(array)) == len(np.unique(ravel(array)))


class TestDatetime64DIndexes:
    @given(
        data=data(),
        n=integers(0, 10),
        valid_dates=booleans(),
        unique=booleans(),
        sort=booleans(),
    )
    def test_main(
        self,
        data: DataObject,
        n: int,
        valid_dates: bool,
        unique: bool,
        sort: bool,
    ) -> None:
        index = data.draw(
            datetime64D_indexes(
                n=n, valid_dates=valid_dates, unique=unique, sort=sort
            )
        )
        assert index.dtype == datetime64D
        assert len(index) == n
        if valid_dates:
            _ = [datetime64_to_date(d) for d in index]
        if unique:
            assert len(index) == len(np.unique(index))
        if sort:
            assert_equal(index, np.sort(index))


class TestDatetime64DTypes:
    @given(dtype=datetime64_dtypes())
    def test_main(self, dtype: Any) -> None:
        _ = dtype


class TestDatetime64Indexes:
    @given(
        data=data(),
        n=integers(0, 10),
        unit=datetime64_units() | none(),
        valid_dates=booleans(),
        valid_datetimes=booleans(),
        unique=booleans(),
        sort=booleans(),
    )
    def test_main(
        self,
        data: DataObject,
        n: int,
        unit: Datetime64Unit | None,
        valid_dates: bool,
        valid_datetimes: bool,
        unique: bool,
        sort: bool,
    ) -> None:
        with assume_does_not_raise(InvalidArgument):
            index = data.draw(
                datetime64_indexes(
                    n=n,
                    unit=unit,
                    valid_dates=valid_dates,
                    valid_datetimes=valid_datetimes,
                    unique=unique,
                    sort=sort,
                )
            )
        assert len(index) == n
        if unit is not None:
            assert datetime64_dtype_to_unit(index.dtype) == unit
        if valid_dates:
            _ = [datetime64_to_date(d) for d in index]
        if valid_datetimes:
            _ = [datetime64_to_datetime(d) for d in index]
        if unique:
            assert len(index) == len(np.unique(index))
        if sort:
            assert_equal(index, np.sort(index))


class TestDatetime64Kinds:
    @given(kind=datetime64_kinds())
    def test_main(self, kind: Datetime64Kind) -> None:
        _ = kind


class TestDatetime64Units:
    @given(data=data(), kind=datetime64_kinds() | none())
    def test_main(self, data: DataObject, kind: Datetime64Kind | None) -> None:
        unit = data.draw(datetime64_units(kind=kind))
        if kind is not None:
            assert datetime64_unit_to_kind(unit) == kind


class TestDatetime64usIndexes:
    @given(
        data=data(),
        n=integers(0, 10),
        valid_datetimes=booleans(),
        unique=booleans(),
        sort=booleans(),
    )
    def test_main(
        self,
        data: DataObject,
        n: int,
        valid_datetimes: bool,
        unique: bool,
        sort: bool,
    ) -> None:
        index = data.draw(
            datetime64us_indexes(
                n=n, valid_datetimes=valid_datetimes, unique=unique, sort=sort
            )
        )
        assert index.dtype == datetime64us
        assert len(index) == n
        if valid_datetimes:
            _ = [datetime64_to_datetime(d) for d in index]
        if unique:
            assert len(index) == len(np.unique(index))
        if sort:
            assert_equal(index, np.sort(index))


class TestDatetime64s:
    @given(data=data(), unit=datetime64_units())
    def test_main(self, data: DataObject, unit: Datetime64Unit) -> None:
        min_value = data.draw(datetime64s(unit=unit) | int64s() | none())
        max_value = data.draw(datetime64s(unit=unit) | int64s() | none())
        with assume_does_not_raise(InvalidArgument):
            datetime = data.draw(
                datetime64s(min_value=min_value, max_value=max_value, unit=unit)
            )
        assert datetime64_dtype_to_unit(datetime.dtype) == unit
        assert not isnat(datetime)
        if min_value is not None:
            if isinstance(min_value, datetime64):
                assert datetime >= min_value
            else:
                assert datetime64_to_int(datetime) >= min_value
        if max_value is not None:
            if isinstance(max_value, datetime64):
                assert datetime <= max_value
            else:
                assert datetime64_to_int(datetime) <= max_value

    @given(
        data=data(),
        min_value=datetime64s(unit="D") | dates() | none(),
        max_value=datetime64s(unit="D") | dates() | none(),
        unit=just("D") | none(),
    )
    def test_valid_dates(
        self,
        data: DataObject,
        min_value: datetime64 | dt.date | None,
        max_value: datetime64 | dt.date | None,
        unit: Literal["D"] | None,
    ) -> None:
        with assume_does_not_raise(InvalidArgument):
            datetime = data.draw(
                datetime64s(
                    min_value=min_value,
                    max_value=max_value,
                    unit=unit,
                    valid_dates=True,
                )
            )
        assert datetime64_dtype_to_unit(datetime.dtype) == "D"
        date = datetime64_to_date(datetime)
        if min_value is not None:
            if isinstance(min_value, datetime64):
                assert datetime >= min_value
            else:
                assert date >= min_value
        if max_value is not None:
            if isinstance(max_value, datetime64):
                assert datetime <= max_value
            else:
                assert date <= max_value

    @given(data=data(), unit=datetime64_units())
    def test_valid_dates_error(
        self, data: DataObject, unit: Datetime64Unit
    ) -> None:
        _ = assume(unit != "D")
        with raises(InvalidArgument):
            _ = data.draw(datetime64s(unit=unit, valid_dates=True))

    @given(
        data=data(),
        min_value=datetime64s(unit="us") | datetimes_utc() | none(),
        max_value=datetime64s(unit="us") | datetimes_utc() | none(),
        unit=just("us") | none(),
    )
    def test_valid_datetimes(
        self,
        data: DataObject,
        min_value: datetime64 | dt.datetime | None,
        max_value: datetime64 | dt.datetime | None,
        unit: Literal["us"] | None,
    ) -> None:
        with assume_does_not_raise(InvalidArgument):
            np_datetime = data.draw(
                datetime64s(
                    min_value=min_value,
                    max_value=max_value,
                    unit=unit,
                    valid_datetimes=True,
                )
            )
        assert datetime64_dtype_to_unit(np_datetime.dtype) == "us"
        py_datetime = datetime64_to_datetime(np_datetime)
        if min_value is not None:
            if isinstance(min_value, datetime64):
                assert np_datetime >= min_value
            else:
                assert py_datetime >= min_value
        if max_value is not None:
            if isinstance(max_value, datetime64):
                assert np_datetime <= max_value
            else:
                assert py_datetime <= max_value

    @given(data=data(), unit=datetime64_units())
    def test_valid_datetimes_error(
        self, data: DataObject, unit: Datetime64Unit
    ) -> None:
        _ = assume(unit != "us")
        with raises(InvalidArgument):
            _ = data.draw(datetime64s(unit=unit, valid_datetimes=True))


class TestFloatArrays:
    @given(
        data=data(),
        shape=array_shapes(),
        min_value=floats() | none(),
        max_value=floats() | none(),
        allow_nan=booleans(),
        allow_inf=booleans(),
        allow_pos_inf=booleans(),
        allow_neg_inf=booleans(),
        integral=booleans(),
        unique=booleans(),
    )
    def test_main(
        self,
        data: DataObject,
        shape: Shape,
        min_value: float | None,
        max_value: float | None,
        allow_nan: bool,
        allow_inf: bool,
        allow_pos_inf: bool,
        allow_neg_inf: bool,
        integral: bool,
        unique: bool,
    ) -> None:
        with assume_does_not_raise(InvalidArgument):
            array = data.draw(
                float_arrays(
                    shape=shape,
                    min_value=min_value,
                    max_value=max_value,
                    allow_nan=allow_nan,
                    allow_inf=allow_inf,
                    allow_pos_inf=allow_pos_inf,
                    allow_neg_inf=allow_neg_inf,
                    integral=integral,
                    unique=unique,
                )
            )
        assert array.dtype == float
        assert array.shape == shape
        if min_value is not None:
            assert (
                (isfinite(array) & (array >= min_value)) | ~isfinite(array)
            ).all()
        if max_value is not None:
            assert (
                (isfinite(array) & (array <= max_value)) | ~isfinite(array)
            ).all()
        if not allow_nan:
            assert (~isnan(array)).all()
        if not allow_inf:
            if not (allow_pos_inf or allow_neg_inf):
                assert (~isinf(array)).all()
            if not allow_pos_inf:
                assert (array != inf).all()
            if not allow_neg_inf:
                assert (array != -inf).all()
        if integral:
            assert ((array == rint(array)) | isnan(array)).all()
        if unique:
            flat = ravel(array)
            assert len(set(flat)) == len(flat)


class TestIntArrays:
    @given(
        data=data(),
        shape=array_shapes(),
        min_value=int64s() | none(),
        max_value=int64s() | none(),
        unique=booleans(),
    )
    def test_main(
        self,
        data: DataObject,
        shape: Shape,
        min_value: int | None,
        max_value: int | None,
        unique: bool,
    ) -> None:
        with assume_does_not_raise(InvalidArgument):
            array = data.draw(
                int_arrays(
                    shape=shape,
                    min_value=min_value,
                    max_value=max_value,
                    unique=unique,
                )
            )
        assert array.dtype == int
        assert array.shape == shape
        if unique:
            flat = ravel(array)
            assert len(set(flat)) == len(flat)


class TestInt32s:
    @given(
        data=data(), min_value=int32s() | none(), max_value=int32s() | none()
    )
    def test_main(
        self, data: DataObject, min_value: int | None, max_value: int | None
    ) -> None:
        with assume_does_not_raise(InvalidArgument):
            x = data.draw(int32s(min_value=min_value, max_value=max_value))
        info = iinfo(int32)
        assert info.min <= x <= info.max
        if min_value is not None:
            assert x >= min_value
        if max_value is not None:
            assert x <= max_value


class TestInt64s:
    @given(
        data=data(), min_value=int64s() | none(), max_value=int64s() | none()
    )
    def test_main(
        self, data: DataObject, min_value: int | None, max_value: int | None
    ) -> None:
        with assume_does_not_raise(InvalidArgument):
            x = data.draw(int64s(min_value=min_value, max_value=max_value))
        info = iinfo(int64)
        assert info.min <= x <= info.max
        if min_value is not None:
            assert x >= min_value
        if max_value is not None:
            assert x <= max_value


class TestStrArrays:
    @given(
        data=data(),
        shape=array_shapes(),
        min_size=integers(0, 100),
        max_size=integers(0, 100) | none(),
        allow_none=booleans(),
        unique=booleans(),
    )
    def test_main(
        self,
        data: DataObject,
        shape: Shape,
        min_size: int,
        max_size: int | None,
        allow_none: bool,
        unique: bool,
    ) -> None:
        with assume_does_not_raise(InvalidArgument):
            array = data.draw(
                str_arrays(
                    shape=shape,
                    min_size=min_size,
                    max_size=max_size,
                    allow_none=allow_none,
                    unique=unique,
                )
            )
        assert array.dtype == object
        assert array.shape == shape
        flat = ravel(array)
        flat_text = [i for i in flat if i is not None]
        assert all(len(t) >= min_size for t in flat_text)
        if max_size is not None:
            assert all(len(t) <= max_size for t in flat_text)
        if not allow_none:
            assert len(flat_text) == array.size
        if unique:
            flat = ravel(array)
            assert len(set(flat)) == len(flat)


class TestUInt32s:
    @given(
        data=data(), min_value=uint32s() | none(), max_value=uint32s() | none()
    )
    def test_main(
        self, data: DataObject, min_value: int | None, max_value: int | None
    ) -> None:
        with assume_does_not_raise(InvalidArgument):
            x = data.draw(uint32s(min_value=min_value, max_value=max_value))
        info = iinfo(uint32)
        assert info.min <= x <= info.max
        if min_value is not None:
            assert x >= min_value
        if max_value is not None:
            assert x <= max_value


class TestUInt64s:
    @given(
        data=data(), min_value=uint64s() | none(), max_value=uint64s() | none()
    )
    def test_main(
        self, data: DataObject, min_value: int | None, max_value: int | None
    ) -> None:
        with assume_does_not_raise(InvalidArgument):
            x = data.draw(uint64s(min_value=min_value, max_value=max_value))
        info = iinfo(uint64)
        assert info.min <= x <= info.max
        if min_value is not None:
            assert x >= min_value
        if max_value is not None:
            assert x <= max_value



from __future__ import annotations

from typing import Any

import sqlalchemy
from click import Context
from click import Parameter
from click import ParamType
from typing_extensions import override

from utilities.sqlalchemy import ParseEngineError
from utilities.sqlalchemy import ensure_engine


class Engine(ParamType):
    """An engine-valued parameter."""

    name = "engine"

    @override
    def convert(
        self, value: Any, param: Parameter | None, ctx: Context | None
    ) -> sqlalchemy.Engine:
        """Convert a value into the `Engine` type."""
        try:
            return ensure_engine(value)
        except ParseEngineError:
            self.fail(f"Unable to parse {value}", param, ctx)



from __future__ import annotations

import datetime as dt
from collections.abc import Hashable
from typing import TYPE_CHECKING
from typing import Any

from hypothesis import assume
from hypothesis.extra.pandas import indexes as _indexes
from hypothesis.strategies import DrawFn
from hypothesis.strategies import SearchStrategy
from hypothesis.strategies import composite
from hypothesis.strategies import dates
from hypothesis.strategies import datetimes
from hypothesis.strategies import integers
from pandas import Timedelta
from pandas import Timestamp

from utilities.datetime import UTC
from utilities.pandas import TIMESTAMP_MAX_AS_DATE
from utilities.pandas import TIMESTAMP_MAX_AS_DATETIME
from utilities.pandas import TIMESTAMP_MIN_AS_DATE
from utilities.pandas import TIMESTAMP_MIN_AS_DATETIME
from utilities.pandas import string

from .hypothesis import MaybeSearchStrategy  # noqa: TID252
from .hypothesis import lift_draw  # noqa: TID252
from .hypothesis import text_ascii  # noqa: TID252
from .numpy import int64s  # noqa: TID252

if TYPE_CHECKING:  # pragma: no cover
    from utilities.pandas import IndexA
    from utilities.pandas import IndexI
    from utilities.pandas import IndexS


@composite
def dates_pd(
    _draw: DrawFn,
    /,
    *,
    min_value: MaybeSearchStrategy[dt.date] = TIMESTAMP_MIN_AS_DATE,
    max_value: MaybeSearchStrategy[dt.date] = TIMESTAMP_MAX_AS_DATE,
) -> dt.date:
    """Strategy for generating dates which can become Timestamps."""
    draw = lift_draw(_draw)
    return draw(dates(min_value=draw(min_value), max_value=draw(max_value)))


@composite
def datetimes_pd(
    _draw: DrawFn,
    /,
    *,
    min_value: MaybeSearchStrategy[dt.datetime] = TIMESTAMP_MIN_AS_DATETIME,
    max_value: MaybeSearchStrategy[dt.datetime] = TIMESTAMP_MAX_AS_DATETIME,
) -> dt.datetime:
    """Strategy for generating datetimes which can become Timestamps."""
    draw = lift_draw(_draw)
    datetime = draw(
        datetimes(
            min_value=draw(min_value).replace(tzinfo=None),
            max_value=draw(max_value).replace(tzinfo=None),
        )
    )
    return datetime.replace(tzinfo=UTC)


_INDEX_LENGTHS = integers(0, 10)


@composite
def indexes(
    _draw: DrawFn,
    /,
    *,
    elements: SearchStrategy[Any] | None = None,
    dtype: Any = None,
    n: MaybeSearchStrategy[int] = _INDEX_LENGTHS,
    unique: MaybeSearchStrategy[bool] = True,
    name: MaybeSearchStrategy[Hashable] = None,
    sort: MaybeSearchStrategy[bool] = False,
) -> IndexA:
    """Strategy for generating Indexes."""
    draw = lift_draw(_draw)
    n_ = draw(n)
    index = draw(
        _indexes(
            elements=elements,
            dtype=dtype,
            min_size=n_,
            max_size=n_,
            unique=draw(unique),
        )
    )
    index = index.rename(draw(name))
    if draw(sort):
        return index.sort_values()
    return index


def int_indexes(
    *,
    n: MaybeSearchStrategy[int] = _INDEX_LENGTHS,
    unique: MaybeSearchStrategy[bool] = True,
    name: MaybeSearchStrategy[Hashable] = None,
    sort: MaybeSearchStrategy[bool] = False,
) -> SearchStrategy[IndexI]:
    """Strategy for generating integer Indexes."""
    return indexes(
        elements=int64s(), dtype=int, n=n, unique=unique, name=name, sort=sort
    )


@composite
def str_indexes(
    _draw: DrawFn,
    /,
    *,
    min_size: MaybeSearchStrategy[int] = 0,
    max_size: MaybeSearchStrategy[int | None] = None,
    n: MaybeSearchStrategy[int] = _INDEX_LENGTHS,
    unique: MaybeSearchStrategy[bool] = True,
    name: MaybeSearchStrategy[Hashable] = None,
    sort: MaybeSearchStrategy[bool] = False,
) -> IndexS:
    """Strategy for generating string Indexes."""
    draw = lift_draw(_draw)
    elements = text_ascii(min_size=min_size, max_size=max_size)
    index = draw(
        indexes(
            elements=elements,
            dtype=object,
            n=n,
            unique=unique,
            name=name,
            sort=sort,
        )
    )
    return index.astype(string)


@composite
def timestamps(
    _draw: DrawFn,
    /,
    *,
    min_value: MaybeSearchStrategy[dt.datetime] = TIMESTAMP_MIN_AS_DATETIME,
    max_value: MaybeSearchStrategy[dt.datetime] = TIMESTAMP_MAX_AS_DATETIME,
    allow_nanoseconds: MaybeSearchStrategy[bool] = False,
) -> Timestamp:
    """Strategy for generating Timestamps."""
    draw = lift_draw(_draw)
    min_value, max_value = (draw(mv) for mv in (min_value, max_value))
    datetime = draw(datetimes_pd(min_value=min_value, max_value=max_value))
    timestamp: Timestamp = Timestamp(datetime)
    if draw(allow_nanoseconds):
        nanoseconds = draw(integers(-999, 999))
        timedelta = Timedelta(nanoseconds=nanoseconds)  # type: ignore
        timestamp += timedelta
        _ = assume(min_value <= timestamp.floor("us"))
        _ = assume(timestamp.ceil("us") <= max_value)
    return timestamp



from __future__ import annotations

from collections.abc import Sequence

from hypothesis import given
from hypothesis.strategies import floats
from hypothesis.strategies import integers
from numpy import array
from numpy import isfinite
from numpy import isnan
from numpy import nan
from numpy.testing import assert_allclose
from numpy.testing import assert_equal
from pytest import mark
from pytest import param

from utilities.hypothesis import float_arrays
from utilities.numpy import NDArrayF1
from utilities.numpy import is_between_or_nan
from utilities.scipy import ppf


class TestPPF:
    @mark.parametrize(
        ("values", "expected"),
        [
            param([], []),
            param([0.0], [0.0]),
            param([0.0, 0.1], [-1.0, 1.0]),
            param([0.0, 0.1, 0.2], [-1.0, 0.0, 1.0]),
            param([0.0, 0.1, 0.2, 0.3], [-1.0, -0.2891889, 0.2891889, 1.0]),
            param([nan], [nan]),
            param([0.0, nan, 0.1], [-1.0, nan, 1.0]),
        ],
    )
    def test_examples(
        self, values: Sequence[float], expected: Sequence[float]
    ) -> None:
        result = ppf(array(values, dtype=float), 1.0)
        assert_allclose(result, array(expected, dtype=float))

    @given(
        array=float_arrays(
            shape=integers(0, 10),
            min_value=-10.0,
            max_value=10.0,
            allow_nan=True,
        ),
        cutoff=floats(0.0, 10.0),
    )
    def test_main(self, array: NDArrayF1, cutoff: float) -> None:
        result = ppf(array, cutoff)
        assert_equal(isfinite(result), isfinite(array))
        assert_equal(isnan(result), isnan(array))
        assert is_between_or_nan(result, -cutoff, cutoff).all()
